 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: base64.c
 *
 * Description: MIME-compliant base 64 encoding library.
 *
 *
 * Global Functions:	
 *	void b64toinit(b64tostate *s);
 *	size_t b64towrite(char *ptr, size_t n, FILE *f, b64tostate *s);
 *	void b64towriteend(FILE *f, b64tostate *s);
 *	void b64fromreadinit(FILE *f, b64fromstate *s);
 *	size_t b64fromread(char *buf, size_t n, FILE *f, b64fromstate *s);
 *	size_t b64fromfgets(char *buf, size_t n, FILE *f, b64fromstate* s);
 *	void b64fromend(b64fromstate *s);
 *
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Alan deLespinasse and Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

/* To create a standalone filter program, compile with the -DSTANDALONE flag.
   The resulting program filters from stding to stdout, encoding if there
   are no arguments, or decoding if there is a "-d" flag. */

#include "base64.h"
#define PUBLIC
#define PRIVATE static
#define FROMBUFSIZE 2000

PRIVATE char to64[] = {
  'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
  'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
  'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
  'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/'
};

PRIVATE short from64[] = {
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63,
  52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-2,-1,-1,
  -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,
  15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,
  -1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,
  41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1
};


/* Fills in a b64tostate to begin encoding. */
PUBLIC void b64toinit(b64tostate *s)
{
  s->phase = s->column = 0;
  s->b = s->p = 0;
}

/* Sends a char to a stream, then sends a newline if the line is long
   enough. */
PRIVATE void put76(char c, FILE *f,int *column)
{
  putc(c,f);
  if (++(*column) == 76) {
    putc('\n',f);
    (*column) = 0;
  }
}

/* Writes n bytes from ptr to f, encoded in base 64. */
PUBLIC size_t b64towrite(char *ptr, size_t n, FILE *f, b64tostate *s)
{
  for (;;) {
    if (!n) break;
    switch (s->phase) {
    case 0:
      s->b = *ptr;
      ptr++;
      n--;
      break;
    case 1:
      s->b = (s->b << 6) | ((*ptr >> 2) & 63);
      s->p = *ptr;
      ptr++;
      n--;
      break;
    case 2:
      s->b = (s->p << 4);
      s->b |= ((*ptr >> 4) & 15);
      break;
    default:
      s->b= *ptr << 2;
      ptr++;
      n--;
    }
    put76(to64[(s->b >> 2) & 63], f, &(s->column));
    s->phase = (s->phase + 1) & 3;
  }
  return(n);
}

/* Writes any leftover bytes to the stream, then an extra newline for luck.
   Does not close the stream. */
PUBLIC void b64towriteend(FILE *f, b64tostate *s)
{
  if (s->phase == 1) {
    put76(to64[(s->b << 4) & 48], f, &(s->column));
    put76('=', f, &(s->column));
    put76('=', f, &(s->column));
  } else if (s->phase == 2) {
    put76(to64[(s->p << 2) & 60], f, &(s->column));
    put76('=', f, &(s->column));
  }
  putc('\n', f);
}


/* Fills in a b64fromstate to start decoding. */
PUBLIC void b64fromreadinit(FILE *f, b64fromstate *s)
{
  s->done = s->shift = s->prevc = s->nextc = 0;
  s->buf = (char*) malloc(FROMBUFSIZE);
  s->bufsize = fread(s->buf, 1, FROMBUFSIZE, f);
  s->bufofs = 0;
}

/* Returns the next 6-bit value from the stream, or a special code: -1 to
   try again, or -2 to signal end of file. */
PRIVATE short getfrombuf(FILE *f, b64fromstate *s)
{
  unsigned char c;

  if (s->bufofs == s->bufsize) {
    s->bufsize = fread(s->buf, 1, FROMBUFSIZE, f);
    if (s->bufsize == 0) return(-2);
    s->bufofs=  0;
  }
  c = s->buf[s->bufofs++];
  if ((c >= 0) && (c <= 127))
    return(from64[(unsigned int)c]);
  else
    return(-1);
}

/* Reads and decodes up to n bytes from f, placing them in buf. Returns 
   number of bytes actually read. */
PUBLIC size_t b64fromread(char *buf, size_t n, FILE *f, b64fromstate *s)
{
  size_t nunc = 0;
  char *bp = buf;

  if (s->done) return(0);

  for (;;) {
    s->nextc = getfrombuf(f, s);
    if (s->nextc == -2) {
      s->done = 1;
      break;
    } else if (s->nextc >= 0) {
      if (s->shift) {
	*bp = (s->prevc << s->shift) | ((char) s->nextc >> (6 - s->shift));
	bp++;
	nunc++;
      }
      s->shift = (s->shift + 2) & 6;
      s->prevc = (char) s->nextc;
    }
    if (nunc == n) break;
  }
  return(nunc);
}

/* Reads and decodes up to n bytes from f, placing them in buf. Stops if
   a newline is read. Returns number of bytes read. */
PUBLIC size_t b64fromfgets(char *buf, size_t n, FILE *f, b64fromstate* s)
{
  size_t nunc = 0;
  char* bp = buf;
  int nl = 0;

  if (s->done) return(0);

  for (;;) {
    s->nextc = getfrombuf(f, s);
    if (s->nextc == -2) {
      s->done = 1;
      break;
    } else if (s->nextc >= 0) {
      if (s->shift) {
	*bp = (s->prevc << s->shift) | ((char) s->nextc >> (6 - s->shift));
	if (*bp == 10) nl = 1;
	bp++;
	nunc++;
      }
      s->shift = (s->shift + 2) & 6;
      s->prevc = (char) s->nextc;
    }
    if (nl || (nunc == n)) {
      *bp = 0;
      break;
    }
  }
  return(nunc);
}

/* Clean up after decoding. */
PUBLIC void b64fromend(b64fromstate *s)
{
  free(s->buf);
}


#ifdef STANDALONE

#define BUFSIZE 5000

PUBLIC int main(int argc, char **argv)
{
  char buf[BUFSIZE];
  size_t n;
  b64tostate tostate;
  b64fromstate fromstate;

  if (argc == 1) {
    b64toinit(&tostate);
    while (n = fread(buf, 1, BUFSIZE, stdin)) {
      b64towrite(buf, n, stdout, &tostate);
    }
    b64towriteend(stdout, &tostate);
  } else if (argc == 2) {
    b64fromreadinit(stdin, &fromstate);
    while (n=b64fromread(buf, BUFSIZE, stdin, &fromstate))
      fwrite(buf, 1, n, stdout);
    b64fromend(&fromstate);
  } else fprintf(stderr, "Usage: base64 [-d]\n");
  return 0;
}

#endif
