 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: database.c
 *
 * Description: Generic log and database interface for GDBM routines.
 *
 *
 * Global Functions: See HTML documentation
 *	
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1996-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define LOGDEBUG

#include "rover.h"
#include "rovertcp.h"
#include "database.h"

PRIVATE char line[2050], tmp[2050];
PRIVATE Tcl_HashTable LogRecords;
PRIVATE int hash_init = 0;

/*
 * Given a database and an object name, returns the object's log
 *   information record. If one does not exist, it is created. The
 *   returned record should never be freed.
 */
PRIVATE inline logInfo *log_getInfo(Database *db, char *obj)
{
  datum info, rq;
  logInfo *rv = NULL;
  Tcl_HashEntry *hentry;
  int new = 0;

  if (!hash_init) {	/* Initialize the hash table, if necessary */
    hash_init = 1;
    Tcl_InitHashTable(&LogRecords, TCL_STRING_KEYS); /* In-memory log htable */
  }
  sprintf(tmp, DB_LOGINFOTEMPLATE, obj);
  
  /* First, check the hash table for the log info record */
  hentry = Tcl_FindHashEntry(&LogRecords, tmp);
  if (hentry) return((logInfo *)Tcl_GetHashValue(hentry));

  /* Now, check the disk db */
  rq.dptr = tmp;
  rq.dsize = strlen(tmp)+1;
  info = gdbm_fetch(db->db, rq);
  if (info.dptr) {
    rv = (logInfo *)(info.dptr);
    rv->dirty = 0;	/* Freshly loaded entries are clean */
  } else {
    /* Otherwise, create a new entry */
    if ((rv = (logInfo *)malloc(sizeof(logInfo))) == NULL) {
      state->status = 500;
      errorExit("Unable to allocate log information");
    }    
    rv->dirty = 1;		/* Freshly created entries are dirty */
    rv->logStart = 0;		/* Number of first log entry */
    rv->logEnd = 0;		/* Number of last log entry */
    rv->logSize = 0;		/* Total number of log records */
    rv->time = 0;
  }
  
  /* Now, insert the entry into the hash table */
  hentry = Tcl_CreateHashEntry(&LogRecords, tmp, &new);
  if (!new)			/* Check for existing entry */
    fprintf(stderr, "Warning! Entry %s appeared in table during create\n",
	    tmp);
  Tcl_SetHashValue(hentry, (char *) rv);	/* Save it */
  return(rv);
}

/*
 * Given a database and an object name, returns the object's log
 *   record. If no record exists, NULL is returned. If opsize is
 *   non-negative, it is compared against the record's operation
 *   size; if they differ, the old record is released and a new
 *   one is created.
 */
PRIVATE inline Log *log_getRecord(Database *db, char *obj, int opsize)
{
  datum info, rq;
  Log *new, *rv = NULL;
  Tcl_HashEntry *hentry;
  int newrec, inhash = 0;

  if (!hash_init) {	/* Initialize the hash table, if necessary */
    hash_init = 1;
    Tcl_InitHashTable(&LogRecords, TCL_STRING_KEYS); /* In-memory log htable */
  }
  /* First, check the hash table for the log record */
  hentry = Tcl_FindHashEntry(&LogRecords, obj);
  if (hentry) {
    inhash = 1;
    rv = (Log *)Tcl_GetHashValue(hentry);
  } else {
    /* Now, check the disk db */
    rq.dptr = obj;
    rq.dsize = strlen(obj)+1;
    info = gdbm_fetch(db->db, rq);
    if (!info.dptr) return(NULL);

    rv = (Log *)(info.dptr);
    rv->dirty = 0;	/* Freshly loaded entries are clean */

    /* Now, insert the entry into the hash table */
    hentry = Tcl_CreateHashEntry(&LogRecords, obj, &newrec);
    if (!newrec)			/* Check for existing entry */
      fprintf(stderr, "Warning! Entry %s appeared in table during create\n",
	      obj);
    Tcl_SetHashValue(hentry, (char *) rv);	/* Save it */
  }
  
  if ((opsize >= 0) && (rv->opsize != opsize)) {
    /* We're changing the log record, create a new entry */
    if ((new = (Log *)malloc(sizeof(Log)+opsize)) == NULL) {
      state->status = 500;
      errorExit("Unable to allocate log record");
    }    
    new->dirty = 1;		/* Freshly created entries are dirty */
    Tcl_SetHashValue(hentry, (char *) new);	/* Save it */
    free(rv);
    rv = new;
  }
  return(rv);
}

/*
 * Save the log record for an object. Will replace an existing entry.
 */
PRIVATE inline void log_saveRecord(Database *db, char *obj, Log *log)
{
  int new;
  Tcl_HashEntry *hentry, *fentry;

  if (!hash_init) {	/* Initialize the hash table, if necessary */
    hash_init = 1;
    Tcl_InitHashTable(&LogRecords, TCL_STRING_KEYS); /* In-memory log htable */
  }
  /* Now, insert the entry into the hash table */
  hentry = Tcl_CreateHashEntry(&LogRecords, obj, &new);
  if (!new) {			/* Check for existing entry */
    fprintf(stderr, "Warning! Entry %s found in table during create\n",
	    obj);
    /* Delete the existing entry */
    fentry = Tcl_FindHashEntry(&LogRecords, obj);
    free(Tcl_GetHashValue(fentry));
    Tcl_DeleteHashEntry(fentry);
  }
  Tcl_SetHashValue(hentry, (char *) log);	/* Save it */
}

/*
 * Flush out all log information and log records.
 */
PUBLIC void log_flush(Database *db)
{
  Tcl_HashSearch search;
  Tcl_HashEntry *hentry;
  int rc, flushed = 0, count = 0, write = db->write, opened;
  char *name, *value;
  datum rq, info;
  logInfo *linfo;
  Log *log;

  if (!hash_init) {	/* If the hash table is not initialized, there are
			   no records to flush. */
#ifdef LOGDEBUG
    fprintf(stderr, "No log records to flush\n");
#endif /* LOGDEBUG */
    return;
  }
  opened = write;
  for (hentry = Tcl_FirstHashEntry(&LogRecords, &search); hentry != NULL;
       hentry = Tcl_NextHashEntry(&search)) {
    /* For each record, get the name and value */
    name = (char *) Tcl_GetHashKey(&LogRecords, hentry);
    value = Tcl_GetHashValue(hentry);
    count++;
    if (!strncmp(name, DB_LOGPREFIX, strlen(DB_LOGPREFIX))) {
      /* Log record */
      log = (Log *) value;
      if (log->dirty) {
	if (!opened) {
	  opened = 1;
	  db_reopenWrite(db);	/* Reopen the db in write mode */
	}
	flushed++;
	rq.dptr = name;
	rq.dsize = strlen(name)+1;
	info.dptr = value;
	info.dsize = sizeof(Log)+log->opsize;
	rc = gdbm_store(db->db, rq, info, GDBM_REPLACE);
	if (rc != 0) {
	  state->status = 500;
	  errorExit("Error while adding log entry");
	}
      }
      free(value);
    } else if (!strncmp(name, DB_LOGINFOPREFIX, strlen(DB_LOGINFOPREFIX))) {
      /* Log information record */
      linfo = (logInfo *) value;
      if (linfo->dirty) {
	if (!opened) {
	  opened = 1;
	  db_reopenWrite(db);	/* Reopen the db in write mode */
	}
	flushed++;
	rq.dptr = name;
	rq.dsize = strlen(name)+1;
	info.dptr = value;
	info.dsize = sizeof(logInfo);
	rc = gdbm_store(db->db, rq, info, GDBM_REPLACE);
	if (rc != 0) {
	  state->status = 500;
	  errorExit("Error while updating log information");
	}
      }
      free(value);
    } else {
      fprintf(stderr, "Found unknown hash table entry `%s'\n", name);
    }
  }

  /* Release all of the space used by the entries */
  Tcl_DeleteHashTable(&LogRecords);
  hash_init = 0;
#ifdef LOGDEBUG
  fprintf(stderr, "Flushed %d log records\n", count);
#endif /* LOGDEBUG */
  if (opened && !write) db_reopenRead(db);
}


/*
 * Purge the log records for an object.
 */   
PUBLIC void log_purge(Database *db, char *obj)
{
  datum rq;
  int write = db->write, i, count = 0, rc;
  logInfo *linfo;
  Tcl_HashEntry *hentry;

  if (!hash_init) {	/* Initialize the hash table, if necessary */
    hash_init = 1;
    Tcl_InitHashTable(&LogRecords, TCL_STRING_KEYS); /* In-memory log htable */
  }
  /* Purge all the entries in the log for an object */
#ifdef LOGDEBUG
  fprintf(stderr, "Purging log for %s\n", obj);
#endif /* LOGDEBUG */
  linfo = log_getInfo(db, obj);

  if (!write) db_reopenWrite(db);	/* Reopen the db in write mode */
  for (i = linfo->logStart; i < linfo->logEnd; i++) {
    sprintf(tmp, DB_LOGTEMPLATE, obj, i);
    hentry = Tcl_FindHashEntry(&LogRecords, tmp);
    if (hentry) {
      free(Tcl_GetHashValue(hentry));
      Tcl_DeleteHashEntry(hentry);
    }
    rq.dptr = tmp;
    rq.dsize = strlen(tmp)+1;
    rc = gdbm_delete(db->db, rq);	/* Delete the log entry*/
    if (rc == 0) count++;
  }
  if (count != linfo->logSize) {
    state->status = 500;
    fprintf(stderr, "Log purge failed: %d v %d\n", count, linfo->logSize);
    errorExit("Log purge failed");
  }
  linfo->logStart = linfo->logEnd = linfo->logSize = 0;
  linfo->time = 0;
  gdbm_reorganize(db->db);		/* Release the disk space */
  linfo->dirty = 1;
}

/*
 * Purge the log records for all objects in the database.
 */   
PUBLIC void log_purgeAll(Database *db)
{
  Tcl_HashSearch search;
  Tcl_HashEntry *hentry;
  datum key, nextkey;
  int i, deleted = 0, count = 0, max = 64, rc, write = db->write;
  char **rec, *name, *value;

  if (!hash_init) {	/* Initialize the hash table, if necessary */
    hash_init = 1;
    Tcl_InitHashTable(&LogRecords, TCL_STRING_KEYS); /* In-memory log htable */
  }
  /* Purge all the log entries */
#ifdef LOGDEBUG
  fprintf(stderr, "Purging all log entries\n");
#endif /* LOGDEBUG */

  if (!write) db_reopenWrite(db);	/* Reopen the db in write mode */
  for (hentry = Tcl_FirstHashEntry(&LogRecords, &search); hentry != NULL;
       hentry = Tcl_NextHashEntry(&search)) {
    /* For each record, get the name and value */
    name = (char *) Tcl_GetHashKey(&LogRecords, hentry);
    value = Tcl_GetHashValue(hentry);
    if (!strncmp(name, DB_LOGPREFIX, strlen(DB_LOGPREFIX))) {
      /* Log record */
      deleted++;
      key.dptr = name;
      key.dsize = strlen(name)+1;
      rc = gdbm_delete(db->db, key);	/* Delete the log entry */
      free(value);
    } else if (!strncmp(name, DB_LOGINFOPREFIX, strlen(DB_LOGINFOPREFIX))) {
      /* Log information record */
      deleted++;
      key.dptr = name;
      key.dsize = strlen(name)+1;
      rc = gdbm_delete(db->db, key);	/* Delete the log entry */
      free(value);
    } else {
      fprintf(stderr, "Found unknown hash table entry %s\n", name);
    }
  }
  
  /* Release all of the space used by the entries */
  Tcl_DeleteHashTable(&LogRecords);
  hash_init = 0;
#ifdef LOGDEBUG
  fprintf(stderr, "Deleted %d hashtable records\n", deleted);
#endif /* LOGDEBUG */

  if ((rec = (char **)malloc(sizeof(char *)*max)) == NULL) {
    fatalError("Out of memory\n");
  } 
  key = gdbm_firstkey(db->db);
  while (key.dptr) {
    if (!strncmp(key.dptr, DB_LOGPREFIX, strlen(DB_LOGPREFIX)) ||
	!strncmp(key.dptr, DB_LOGINFOPREFIX, strlen(DB_LOGINFOPREFIX))) {
      if ((count+1) >= max) {
	max <<= 1;
	if ((rec = (char **)realloc(rec,sizeof(char *)*max)) == NULL) {
	  fatalError("Out of memory\n");
	}
      }
      rec[count++] = key.dptr;
    } else 
      free(key.dptr);
    nextkey = gdbm_nextkey(db->db, key);
    key = nextkey;
  }

  for (i = 0; i < count; i++) {
    key.dptr = rec[i];
    key.dsize = strlen(rec[i])+1;
    rc = gdbm_delete(db->db, key);	/* Delete the log entry*/
  }
  free(rec);
#ifdef LOGDEBUG
  fprintf(stderr, "Deleted %d db records\n", count);
#endif /* LOGDEBUG */
  gdbm_reorganize(db->db);		/* Release the disk space */
  if (!write) db_reopenRead(db);
}

/*
 * Clean the log records for an object in the database.
 */
PUBLIC void log_clean(Database *db, char *obj, LogCleaner cleaner)
{
  int rc, i, max = 0, delete, count = 0, dirty = 0;
  datum rq;
  logInfo *linfo;
  time_t time = 0;
  Tcl_HashEntry *hentry;
  Log *log;

  if (!hash_init) {	/* Initialize the hash table, if necessary */
    hash_init = 1;
    Tcl_InitHashTable(&LogRecords, TCL_STRING_KEYS); /* In-memory log htable */
  }
  /* Clean the log by invoking the specified cleaner function for each log
     entry. If the cleaner function returns a non-zero result, delete the log
     entry.
   */
  linfo = log_getInfo(db, obj);
  for (i = linfo->logStart; i < linfo->logEnd; i++) {
    sprintf(tmp, DB_LOGTEMPLATE, obj, i);
    log = log_getRecord(db, tmp, -1);
    if (log == NULL) {
      if (i == linfo->logStart) linfo->logStart++;
      dirty = 1;
      continue;
    }
    delete = cleaner(log);
    if (delete) {
      dirty = 1;
      linfo->logSize--;
      sprintf(tmp, DB_LOGTEMPLATE, obj, i);
      hentry = Tcl_FindHashEntry(&LogRecords, tmp);
      if (hentry) {
	free(Tcl_GetHashValue(hentry));
	Tcl_DeleteHashEntry(hentry);
      }

      rq.dptr = tmp;
      rq.dsize = strlen(tmp)+1;
      rc = gdbm_delete(db->db, rq);	/* Delete the log entry*/
      if (i == linfo->logStart) linfo->logStart++;
    } else {
      time = log->time;
      max = i;
      count++;
    }
  }
  if (count != linfo->logSize) {
    state->status = 500;
    fprintf(stderr, "Log clean failed: %d v %d\n", count, linfo->logSize);
    errorExit("Log clean failed");
  }
  linfo->logEnd = max;
  linfo->time = time;
  gdbm_reorganize(db->db);		/* Release any freed space */
  if (dirty) linfo->dirty = 1;
}

/*
 * Add an log record for an object in the data base.
 */
PUBLIC void log_add(Database *db, char *obj, time_t stamp, char *host,
		    int logID, RoverOpStatus status, char *operation, int opsize)
{
  Log *log;
  logInfo *linfo;
  int new = 0;

  linfo = log_getInfo(db, obj);
  sprintf(tmp, DB_LOGTEMPLATE, obj, linfo->logEnd);
  log = log_getRecord(db, tmp, opsize);
  if (!log) {
    new = 1;
    if ((log = (Log *) malloc(sizeof(Log)+opsize)) == NULL) {
      state->status = 500;
      errorExit("Unable to allocate log entry");
    }
  }
  strcpy(log->clientHost, host);
  log->logID = logID;
  log->status = status;
  linfo->time = log->time = stamp;
  log->opsize = opsize;
  strcpy(log->operation, operation);
  if (new) log_saveRecord(db, tmp, log);
  linfo->logSize++;
  linfo->logEnd++;
  linfo->dirty = 1;
}

/*
 * Search for a matching log record for an object in the database.
 */
PUBLIC Log *log_search(Database *db, char *obj, char *host, int logID)
{
  int i;
  Log *log;
  logInfo *linfo;

  linfo = log_getInfo(db, obj);
  for (i = linfo->logStart; i < linfo->logEnd; i++) {
    sprintf(tmp, DB_LOGTEMPLATE, obj, i);
    log = log_getRecord(db, tmp, -1);
    if (log == NULL) continue;		/* Avoid marking linfo as dirty */
    if (!strcmp(host, log->clientHost) && (logID == log->logID)) return(log);
  }
  return(NULL);
}

/*
 * Print the log records for an object in the database.
 */
PUBLIC void log_print(Database *db, char *obj, LogPrinter printer, FILE *file)
{
  int i;
  char *type, *text;
  Log *log;
  logInfo *linfo;

  linfo = log_getInfo(db, obj);
  fprintf(file, "Log - %s: Start %d, End %d, Size %d Time %u\n", obj,
	  linfo->logStart, linfo->logEnd, linfo->logSize,
	  (unsigned int)linfo->time);
  for (i = linfo->logStart; i < linfo->logEnd; i++) {
    sprintf(tmp, DB_LOGTEMPLATE, obj, i);
    log = log_getRecord(db, tmp, -1);
    if (log == NULL) continue;		/* Avoid marking linfo as dirty */
    switch (log->status) {
    case OP_COMMIT:		/* Operation committed */
      type = "OP_COMMIT";
      break;
    case OP_ABORT:		/* Operation aborted */
      type = "OP_ABORT";
      break;
    default:
      type = "unknown";
      break;
    }
    text = printer(log);
    fprintf(file, 
	    "%d - %s: Client %s LogID %d Time %lu Opsize %d Operation %s\n",
	    i, type, log->clientHost, log->logID, log->time, log->opsize,text);
    free(text);
  }
}

/*
 * Given a log record, output it.
 */
PUBLIC int log_outstrSuffix(Log *log)
{
#ifdef LOGDEBUG
  fprintf(stderr, "LS: %lu %s %d %d `%s'\n", (unsigned long) log->time,
	  log->clientHost, log->logID, (int) log->status, log->operation);
#endif /* LOGDEBUG */
  outstrL(log->time, log->clientHost, log->logID, log->status, log->operation);
  return(1);
}

/*
 * Generate a log suffix for an object in the database.
 */
PUBLIC void log_suffix(Database *db, char *obj, time_t start,
		       LogUpdater updater)
{
  int i, count = 0, rc = 0;
  Log *log;
  logInfo *linfo;

  linfo = log_getInfo(db, obj);
#ifdef LOGDEBUG
  fprintf(stderr,
	  "Log suffix - %s: Start %d, End %d, Size %d; Time %lu; Start %lu\n",
	  obj, linfo->logStart, linfo->logEnd, linfo->logSize, linfo->time,
	  start);
#endif /* LOGDEBUG */
  if (linfo->time > start) {
    for (i = linfo->logStart; i < linfo->logEnd; i++) {
      sprintf(tmp, DB_LOGTEMPLATE, obj, i);
      log = log_getRecord(db, tmp, -1);
      if (log == NULL) continue;	/* Avoid marking linfo as dirty */
      if (log->time <= start) continue;
      rc = updater(log);
      if (rc) count++;
    }
  }
#ifdef LOGDEBUG
  fprintf(stderr, "Sending %d log entries for %s\n", count, obj);
#endif /* LOGDEBUG */
}  

/*
 * Return information about an object's log records.
 */
PUBLIC int log_info(Database *db, char *obj, time_t *start, time_t *end,
		    int *size, time_t *time)
{
  logInfo *linfo = log_getInfo(db, obj);

  if (!linfo) return(-1);
  *start = linfo->logStart;
  *end = linfo->logEnd;
  *size = linfo->logSize;
  *time = linfo->time;
#ifdef LOGDEBUG
  fprintf(stderr,
	  "Log suffix - %s: Start %d, End %d, Size %d; Time %lu\n",
	  obj, linfo->logStart, linfo->logEnd, linfo->logSize, linfo->time);
#endif /* LOGDEBUG */
  return(1);
}  


/*
 * Open a database for writing.
 */
PUBLIC void db_openWrite(char *filename, Database *db, DbBuilder builder,
			 mode_t mode)
{
  int retry = 360;	/* Retry for one minute */
  struct stat sbuf;

#ifdef LOGDEBUG
  fprintf(stderr, "Opening db `%s' in write mode\n", filename);
#endif /* LOGDEBUG */
  ((Database *) state->db) = db;
  db->filename = copyString(filename);
  db->mode = mode;
  if (stat(db->filename, &sbuf))
    db->newdb = 1;
  else
    db->newdb = 0;
  while (--retry) {
    if (((db->db= gdbm_open(db->filename, 0, GDBM_WRCREAT, db->mode,errorExit))
	!= NULL) ||
	(gdbm_errno != GDBM_CANT_BE_WRITER)) {
	break;
    }
    sleep(1);
  }
  if (db->db == NULL) {
    state->status = 400;
    sprintf(tmp, "Rover db write open failed: %s\n",gdbm_strerror(gdbm_errno));
    errorExit(tmp);
  }
  db->write = 1;
  if (db->newdb) {
#ifdef LOGDEBUG
    fprintf(stderr, "New db, recreating\n");
#endif /* LOGDEBUG */
    if (builder) builder(db);
  }
}
  
/*
 * Open a database for reading.
 */
PUBLIC void db_openRead(char *filename, Database *db, DbBuilder builder,
			mode_t mode)
{
  int retry = 360;	/* Retry for six minutes */
  struct stat sbuf;

#ifdef LOGDEBUG
  fprintf(stderr, "Opening db `%s' in read mode\n", filename);
#endif /* LOGDEBUG */
  ((Database *) state->db) = db;
  db->filename = copyString(filename);
  db->mode = mode;
  if (stat(db->filename, &sbuf)) {
    db->newdb = 1;
    db_openWrite(filename, db, builder, mode);
    db_reopenRead(db);
    return;
  }

  db->newdb = 0;
  while (--retry) {
    if (((db->db= gdbm_open(db->filename, 0, GDBM_READER, mode, errorExit))
	 != NULL) ||
	(gdbm_errno != GDBM_CANT_BE_READER)) {
	break;
    }
    sleep(1);
  }

  if (db->db == NULL) {
    state->status = 400;
    sprintf(tmp, "Rover db read open failed: %s\n", gdbm_strerror(gdbm_errno));
    errorExit(tmp);
  }
  db->write = 0;
}
  
/*
 * Reopen the database in write mode.
 */
PUBLIC void db_reopenWrite(Database *db)
{
  int retry = 360;	/* Retry for one minute */

  if (db->write) {
    state->status = 400;
    errorExit("Tried to reopen-write already writeable db");
  }
  gdbm_close(db->db);	/* First close the db */
#ifdef LOGDEBUG
  fprintf(stderr, "Re-opening db `%s' in write mode\n", db->filename);
#endif /* LOGDEBUG */
  while (--retry) {
    if (((db->db= gdbm_open(db->filename, 0, GDBM_WRCREAT, db->mode,errorExit))
	!= NULL) ||
	(gdbm_errno != GDBM_CANT_BE_WRITER)) {
	break;
    }
    sleep(1);
  }
  if (db == NULL) {
    state->status = 400;
    sprintf(tmp, "Rover db reopen-write failed: %s\n",
	    gdbm_strerror(gdbm_errno));
    errorExit(tmp);
  }
  db->write = 1;
}
  
/*
 * Reopen the database in read mode.
 */
PUBLIC void db_reopenRead(Database *db)
{
  int retry = 360;	/* Retry for one minute */

  if (!db->write) {
    state->status = 400;
    errorExit("Tried to reopen-read already read-only db");
  }
  gdbm_close(db->db);	/* First close the db */
#ifdef LOGDEBUG
  fprintf(stderr, "Re-opening db `%s' in read mode\n", db->filename);
#endif /* LOGDEBUG */
  while (--retry) {
    if (((db->db= gdbm_open(db->filename, 0, GDBM_READER, db->mode, errorExit))
	 != NULL) ||
	(gdbm_errno != GDBM_CANT_BE_READER)) {
	break;
    }
    sleep(1);
  }

  if (db->db == NULL) {
    state->status = 400;
    sprintf(tmp,"Rover db reopen-read failed: %s\n",gdbm_strerror(gdbm_errno));
    errorExit(tmp);
  }
  db->write = 0;
}
  
/*
 * Close the database.
 */
PUBLIC void db_close(Database *db)
{
  if (!db || !db->db) return;
  log_flush(db);
  gdbm_close(db->db);
  free(db->filename);
  db->db = NULL;
  ((Database *) state->db) = NULL;
}

/* 
 * Generate a unique object identifier. 
 *  Template takes a single %ld
 */
PUBLIC char *db_getUID(Database *db, char *UIDtemplate)
{
  datum rq;
  struct timeval tv;
  struct timezone tz;
  char *buf;

  if ((buf = malloc(strlen(UIDtemplate)+32)) == NULL) {
    state->status = 500;
    errorExit("Unable to allocate temporary buffer");
  }
  do {
    gettimeofday(&tv, &tz);
    sprintf(buf, UIDtemplate, tv.tv_sec);
    rq.dptr = buf;
    rq.dsize = strlen(buf)+1;
    if (!gdbm_exists(db->db, rq)) {
      datum info;
      int rc;
      /* Create an entry, so concurrent updates won't collide */
      info.dptr = rq.dptr;
      info.dsize = rq.dsize;
      rc = gdbm_store(db->db, rq, info, GDBM_INSERT);
      if (rc != 0) continue;
#ifdef LOGDEBUG
      fprintf(stderr, "Found UID: %s\n", buf);
#endif /* LOGDEBUG */
      break;
    }
  } while (1);	  /* Loop until we create a unique id */
  return(buf);
}  

/*
 * Set the access time for an object in the database.
 */
PUBLIC void db_setAccessTime(Database *db, char *obj, time_t stamp)
{
  int rc, write = db->write;
  datum info, rq;
  
#ifdef LOGDEBUG
  fprintf(stderr, "Set time for %s to %lu\n", obj, (unsigned long) stamp);
#endif /* LOGDEBUG */
  if (!write) db_reopenWrite(db);
  sprintf(line, DB_ACCESSTEMPLATE, obj, state->rhost, state->ClientUser);
  rq.dptr = line;
  rq.dsize = strlen(line)+1;
  info.dptr = (char *) &stamp;
  info.dsize = sizeof(time_t);
  rc = gdbm_store(db->db, rq, info, GDBM_REPLACE);
  if (rc != 0) {
    state->status = 500;
    sprintf(line, "Error while setting access time for %s", obj);
    errorExit(line);
  }
  if (!write) db_reopenRead(db);
}

/*
 * Get the access time for an object in the database.
 */
PUBLIC time_t db_getAccessTime(Database *db, char *obj)
{
  time_t stamp = (time_t) 0;
  datum info, rq;
  
  sprintf(line, DB_ACCESSTEMPLATE, obj, state->rhost, state->ClientUser);
  rq.dptr = line;
  rq.dsize = strlen(line)+1;
  info = gdbm_fetch(db->db, rq);
  if (info.dptr == NULL) return((time_t) 0);
  stamp = *((time_t *)(info.dptr));
  free(info.dptr);
#ifdef LOGDEBUG
  fprintf(stderr, "Get time for %s is %lu\n", obj, (unsigned long) stamp);
#endif /* LOGDEBUG */
  return(stamp);
}
