 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: database.h
 *
 * Description: Generic log and database interface for GDBM routines.
 *
 *
 * Global Functions: None
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef DATABASE_H
#define DATABASE_H

#include <sys/time.h>
#include <gdbm.h>

struct _log {
  int dirty;		/* Record is dirty and should be flushed */
  char clientHost[64];	/* Client's  hostname */
  int logID;		/* Operation's log ID */
  RoverOpStatus status;	/* Entry status */
  time_t time;		/* Entry time */
  int opsize;		/* Size of the operation */
  char operation[0];	/* The operation */
};
typedef struct _log Log;

struct _logInfo {
  int dirty;		/* Record is dirty and should be flushed */
  int logStart;		/* Number of first log entry */
  int logEnd;		/* Number of last log entry */
  int logSize;		/* Total number of log records */
  time_t time;		/* Timestamp of latest entry */
};
typedef struct _logInfo logInfo;

struct _Database {
  int write;		/* Db opened for write? */
  int newdb;		/* New database */
  GDBM_FILE db;		/* GDBM db pointer */
  char *filename;	/* Database filename */
  mode_t mode;		/* Database file mode */
  void *priv;	/* Private application information */
};
typedef struct _Database Database;


#undef DB_VERSION
#define DB_VERSION 		2
#define DB_LOGTEMPLATE		"Rover__Log__%s__%d"
#define DB_LOGPREFIX		"Rover__Log__"
#define DB_LOGINFOTEMPLATE	"Rover__LogInfo__%s"
#define DB_LOGINFOPREFIX	"Rover__LogInfo__"
#define DB_ACCESSTEMPLATE	"Rover__ATime__%s__%s__%s"

typedef int (*LogCleaner)(Log *log);
typedef char *(*LogPrinter)(Log *log);
typedef int (*LogUpdater)(Log *log);
typedef void (*DbBuilder)(Database *db);

/* Log manipulation functions */

extern void log_purge(Database *db, char *obj);
extern void log_purgeAll(Database *db);
extern void log_clean(Database *db, char *obj, LogCleaner cleaner);
extern void log_add(Database *db, char *obj, time_t stamp, char *host,
		    int logID, RoverOpStatus status, char *operation, int opsize);
extern Log *log_search(Database *db, char *obj, char *host, int logID);
extern void log_print(Database *db, char *obj, LogPrinter printer, FILE *file);
extern int log_outstrSuffix(Log *log);
extern void log_suffix(Database *db, char *obj, time_t start,
		       LogUpdater updater);
extern int log_info(Database *db, char *obj, time_t *start, time_t *end,
		    int *size, time_t *time);

/* Database functions */

extern void db_openWrite(char *filename, Database *db, DbBuilder builder,
			 mode_t mode);
extern void db_openRead(char *filename, Database *db, DbBuilder builder,
			mode_t mode);
extern void db_reopenWrite(Database *db);
extern void db_reopenRead(Database *db);
extern void db_close(Database *db);
extern void db_setAccessTime(Database *db, char *obj, time_t stamp);
extern time_t db_getAccessTime(Database *db, char *obj);
extern char *db_getUID(Database *db, char *UIDtemplate);

#endif /* DATABASE_H */
