/**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rmgate.c
 *
 * Description: Server (gateway) program for Rover HTTP proxy.
 *
 *
 * Global Functions:
 *    main
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Alan deLespinasse and Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include <sys/types.h>
#ifndef linux
#include <errno.h>
#include <varargs.h>
#else
#include <stdarg.h>
#include <unistd.h>
#endif /* !linux */
#include <tcl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "base64.h"
#include "html.h"
#include "zlib.h"
#include "RoverTCP.h"

extern int openTCP(char *host, int port);
extern void sendMailHeader(FILE *file, char *url, char *origUrl,int docompress,
		    int image, char *rhost, int rport, int logID, char *source,
		    char *sourcetitle, char *path, char *user, int form,
		    unsigned long time1, unsigned long time2);
static Tcl_HashTable hash_table;/* Keeps track of URLs already fetched. */

int RoverDebug = 0;
char *mailcmd;			/* command line for e-mail program */
char *wwwcmd;			/* command line for line-mode browser */

/* All the data sent from the access manager in the MIME headers. It has
   the same names here as in the access manager (HTRoverHTTP.c)  */
char *source=NULL, *sourcetitle=NULL, *path=NULL, *user=NULL;
char *reply=NULL, *method=NULL, *url=NULL, *rhost=NULL, *origUrl = NULL;
int logID, rport=8080, form=0, prefetch=0, doinline=1, docompress=0;
time_t time1=0x7FFFFFFF,time2=0;
long preftime=3600;


/* Dummy definitions */
char *buffer;


/* These structures implement a data structure built when HTML is parsed
   for other URLs to prefetch. This first one is for a simple linked list
   of URLs to fetch. */
typedef struct prefetchpage {
  char *url;
  struct prefetchpage* next;
} prefetchpage;

/* And this one is the header for the above list. It's passed to the parsing
   callback functions (as the extra void* in the HTML library-- see html.h)
   so they know where to add the new list member, how to interpret relative
   URLs, etc. */
typedef struct parsedata {
  int prefetch,doinline;
  char *source,			/* page name of current fetch job */
  *base;			/* URL of same */
  struct prefetchpage *pagetop,*pagebottom; /* of above linked list */
} parsedata;

/* Defined below (we've got some recursion going on here). */
char* fetch(char*,char*,int,char*,char*,char*,int,time_t,time_t,
	    FILE*,int,int,char*);

/* Dump everything we know to the error log and exit. */
void puke(char* msg) {
  char s[1024];

  fprintf(stderr,ctime(&time2));
  fprintf(stderr,"Fatal error: %s\n",msg);
  fprintf(stderr,"Hostname: %s\n",rhost);
  fprintf(stderr,"Port: %d\n",rport);
  fprintf(stderr,"URL: %s\n",url);
  fprintf(stderr,"Log ID: %d\n", logID);
  fprintf(stderr,"Source: %s\n",source);
  fprintf(stderr,"Sourcetitle: %s\n",sourcetitle);
  fprintf(stderr,"Path: %s\n",path);
  fprintf(stderr,"User: %s\n",user);
  fprintf(stderr,"Reply: %s\n",reply);
  fprintf(stderr,"Form: %s\n",form ? "yes" : "no");
  fprintf(stderr,"Time1: %lu\n",(unsigned long) time1);
  fprintf(stderr,"Time2: %lu\n",(unsigned long) time2);
  fprintf(stderr,"Method: %s\n",method);
  fprintf(stderr,"Prefetch: %d\n",prefetch);
  fprintf(stderr,"Preftime: %ld\n",preftime);
  fprintf(stderr,"Inline: %s\n",doinline ? "yes" : "no");
  fprintf(stderr,"Compress: %d\n",docompress);
  fprintf(stderr,"Remainder of stdin:\n");
  while (fgets(s,1000,stdin)) fputs(s,stderr);
  fprintf(stderr,"\n\n");
  fflush(stderr);
  exit(0);
}

/* Dump everything we know to the error log. */
void info(char* msg) {
  char s[1024];

  fprintf(stderr,ctime(&time2));
  fprintf(stderr,"Info: %s\n",msg);
  fprintf(stderr,"Hostname: %s\n",rhost);
  fprintf(stderr,"Port: %d\n",rport);
  fprintf(stderr,"URL: %s\n",url);
  fprintf(stderr,"Log ID: %d\n", logID);
  fprintf(stderr,"Source: %s\n",source);
  fprintf(stderr,"Sourcetitle: %s\n",sourcetitle);
  fprintf(stderr,"Path: %s\n",path);
  fprintf(stderr,"User: %s\n",user);
  fprintf(stderr,"Reply: %s\n",reply);
  fprintf(stderr,"Form: %s\n",form ? "yes" : "no");
  fprintf(stderr,"Time1: %lu\n",(unsigned long) time1);
  fprintf(stderr,"Time2: %lu\n",(unsigned long) time2);
  fprintf(stderr,"Method: %s\n",method);
  fprintf(stderr,"Prefetch: %d\n",prefetch);
  fprintf(stderr,"Preftime: %ld\n",preftime);
  fprintf(stderr,"Inline: %s\n",doinline ? "yes" : "no");
  fprintf(stderr,"Compress: %d\n",docompress);
  fprintf(stderr,"Remainder of stdin:\n");
  while (fgets(s,1000,stdin)) fputs(s,stderr);
  fprintf(stderr,"\n\n");
  fflush(stderr);
}

/* The callback function for parsing HTML. Goes and fetches inlined images
   right away, so they hopefully get there before the main document.
   Documents to prefetch are stored in a list for later fetching. Does
   not handle image maps correctly. */
void tagcallback(struct HTMLparser* p)
{
  struct parsedata* pd=(struct parsedata*) p->data;
  char* t=strtok(p->buf," <>=\t\n\r");
  if (strcasecmp(t,"a")==0) {	/* it's a hypertext link */
    struct prefetchpage* pfp;

    if (pd->prefetch<=0) return; /* Reached limit of recursion */
    for (;;) {
      t=strtok(NULL," <>=\t\n\r"); /* find the HREF */
      if (t==NULL) return;
      if (strcasecmp(t,"href")==0) break;
    }
    t=strtok(NULL," <>=\"\t\n\r"); /* find the URL */
    if (t==NULL) return;

    pfp=(struct prefetchpage*) malloc(sizeof(struct prefetchpage));
    pfp->url=strdup(t);
    pfp->next=NULL;		/* add the URL to the */
    if (pd->pagetop) {		/* bottom of the list */
      pd->pagebottom->next=pfp;
      pd->pagebottom=pfp;
    }
    else pd->pagetop=pd->pagebottom=pfp;
  }
  else if (strcasecmp(t,"img")==0 || strcasecmp(t,"image")==0 ||
	   strcasecmp(t,"fig")==0) { /* it's an inlined image */
    time_t tt;

    if (!pd->doinline) return;
    for (;;) {
      t=strtok(NULL," <>=\t\n\r"); /* find the SRC reference */
      if (t==NULL) return;
      if (strcasecmp(t,"src")==0) break;
    }
    t=strtok(NULL," <>=\"\t\n\r"); /* find the URL */
    if (t==NULL) return;
    tt=time(NULL);
    fetch(t,"GET",1,pd->base,NULL,NULL,0,tt,tt,NULL,0,0,pd->base);
  }
}

/* Fetch a document using HTTP, and don't give up easily. Send it back
   using mailcmd. Parse HTML, if necessary, and do prefetching stuff.
   Returns an error message string; NULL on success. */
char* fetch_http(char* url,char* method,int image,char* hostname,int port,
		 char* urlpath,char* source,char* sourcetitle,char* path,
		 int form,time_t time1,time_t time2,FILE* request,int prefetch,
		 int doinline)
{
  FILE *sockin,*sockout,*mailout;
  b64tostate state;
  struct parsedata pd;
  struct prefetchpage* pfp;
  int sock, n, linebeg, respcode, html = 0, redirect = 0;
  char s[1024], *line, *type = NULL, *dest = NULL;
  struct HTMLparser* parser = NULL;

  if ((sock = openTCP(hostname,port)) == -1) return "openTCP failed.";
  fprintf(stderr, "open\n");
  sockin = fdopen(sock,"r");	/* Convert to streams. */
  sockout=fdopen(sock,"w");
  if (!sockin || !sockout) return "dfopen(sock) failed.";

  fprintf(sockout,"%s %s HTTP/1.0\n",method,urlpath); /* request */
  if (request)
    while (!feof(request) && ((n=fread(s,1,1000,request))!= 0))
      fwrite(s,1,n,sockout);
  else
    putc('\n',sockout);

  fflush(sockout);

  fgets(s,1000,sockin);
  n = sscanf(s,"HTTP/1.0 %d",&respcode);
  if (!n) respcode = 0;		/* For obsolete 0.9 servers */
  fprintf(stderr, "Response is `%s'\n", s);
  if (respcode==301 || respcode==302) { /* been redirected */
    redirect = 1;
  }

  mailout = popen(mailcmd,"w");	/* send reply via e-mail */
  if (!mailout) return "popen(mailcmd) failed.";

  sendMailHeader(mailout, url, origUrl, docompress, image, rhost, rport, logID,
		 source, sourcetitle, path, user, form, time1, time2);

  sendMailHeader(stderr, url, origUrl, docompress, image, rhost, rport, logID,
		 source, sourcetitle, path, user, form, time1, time2);

  b64toinit(&state);		/* send response in base 64 */
  b64towrite(s,strlen(s),mailout,&state);

  if (strncasecmp(url,"http:",5)==0) { /* only HTTP uses MIME headers */
    for (;!feof(sockin);) {
      fgets(s,1000,sockin);
      b64towrite(s,strlen(s),mailout,&state); /* process headers */
      linebeg=strspn(s," \t\r\n"); /* and parse for content-type */
      if (linebeg==strlen(s)) break;
      line=&s[linebeg];
      if (strncasecmp(line,"content-type:",13)==0)
	type=strdup(strtok(&line[13]," \t\n\r"));
      if (!redirect) continue;

      if (strncasecmp(line,"location:",9)==0) dest=strtok(&s[9]," \t\n\r");
      else if  (strncasecmp(line,"url:",4)==0) dest=strtok(&s[4]," <>\t\n\r");
      if (dest) {
	line = fetch(dest,"GET",image,source,sourcetitle,NULL,form,
		     time1,time2,NULL,prefetch,doinline,url);
	if (line) return(line);
	redirect = 0;
      }
    }
  }
  /* This should probably also check for "application/html", etc. */
  if (prefetch || doinline) {
    if (type) { if (strcasecmp(type,"text/html")==0) html=1; }
    else { if (strcasecmp(strrchr(url,'.'),".html")==0) html=1; }
  }
  if (html) {			/* then parse HTML */
    pd.prefetch=prefetch;
    pd.doinline=doinline;
    pd.source=source;
    pd.base=url;
    pd.pagetop=pd.pagebottom=NULL;
    parser=HTMLnewparse(NULL,tagcallback,(void *) &pd);
  }
  while (!feof(sockin) && ((n=fread(s,1,1000,sockin)) != 0 )) {
    /* Do the whole message body */
    b64towrite(s,n,mailout,&state);
    if (html) HTMLparse(parser,s,n);
  }
  if (html) HTMLendparse(parser);
  b64towriteend(mailout,&state);
  pclose(mailout);
  fclose(sockin);
  fclose(sockout);
  close(sock);

  if (html && prefetch) {
    for (pfp=pd.pagetop; pfp; pfp=pfp->next) {
      fetch(pfp->url,"GET",0,url,NULL,NULL,0,
	    time2,time2,NULL,prefetch-1,doinline,url);
    }
  }

  return NULL;
}

/* Fetch a document by calling the line-mode browser, www. Could work
   for non-form HTTP fetches, but the above function is more efficient
   and does all the extra prefetching stuff that this one doesn't, and
   can also do forms. This function is only used for non-HTTP fetches.
   Returns an error message string; NULL on success. */
char* fetch_other(char *url,char* source, int image,
		char *sourcetitle,char* path,int form,
		time_t time1,time_t time2) {
  FILE *mailout,*wwwin;
  char s[1024];
  b64tostate state;
  int n;

  wwwin=popen(wwwcmd,"r");
  if (wwwin==NULL) return "popen(wwwcmd) failed.";

  mailout=popen(mailcmd,"w");
  if (mailout==NULL) return "popen(mailcmd) failed.";

  sendMailHeader(mailout, url, origUrl, docompress, image, rhost, rport, logID,
		 source, sourcetitle, path, user, form, time1, time2);

  b64toinit(&state);
/*  b64towrite("MIME/1.0 200 Ok\n\n",16,mailout,&state);*/
  while (!feof(wwwin) && ((n=fread(s,1,1000,wwwin)) != 0))
    b64towrite(s,n,mailout,&state);
  b64towriteend(mailout,&state);

  pclose(mailout);
  pclose(wwwin);
  return NULL;
}

/* Fetch a URL. If the URL is relative, turns it into absolute. Doesn't
   fetch if the same URL has already been fetched. The stream pointer
   can optionally contain the rest of the request message after the
   first line; use NULL if it doesn't. */
char* fetch(char *url,char* method,int image,char* source,
	    char* sourcetitle,char* path,int form,time_t time1,
	    time_t time2,FILE* f,int prefetch,int doinline,char *base) {
  char *es,*protocol,*hostname,*urlpath;
  int port, new = 0;
  Tcl_HashEntry *hentry;

  if ((es=parseURL(&url,&protocol,&hostname,&port,&urlpath,base)) != NULL)
      return es;
  if ((hentry = Tcl_FindHashEntry(&hash_table, url)) != NULL)
    return NULL;
  hentry = Tcl_CreateHashEntry(&hash_table, url, &new);
  if (!new) {				/* Check for duplicate urls */
    fprintf(stderr, "Found duplicate url: %s\n", url);
  }
  Tcl_SetHashValue(hentry, (char *) NULL);

  if (strcasecmp(protocol,"http")==0)
    return fetch_http(url,method,image,hostname,port,urlpath,source,
		      sourcetitle,path,form,time1,time2,f,prefetch,
		      doinline);
  else
    return fetch_other(url,source,image,
		       sourcetitle,path,form,time1,time2);
}

int main(int argc,char** argv) {
  char s[1024],*es;
  
  time2=time(NULL);
  Tcl_InitHashTable(&hash_table, TCL_STRING_KEYS);
  freopen("rmgate_log","a",stderr); /* redirect errors to log */
  
  /* Parse SMTP headers. */
  for (;;) {
    if (!fgets(s,1000,stdin)) puke("No header/body separator.");
    if (strspn(s," \t\n\r")==strlen(s)) break;
    
    if (strncasecmp(s,"x-url:",6)==0)
      origUrl=url=strdup(strtok(&s[6]," \t\n\r"));
    else if (strncasecmp(s,"x-hostname: ",12)==0)
      rhost=strdup(strtok(&s[12]," \t\n\r"));
    else if (strncasecmp(s,"x-port:",7)==0)
      rport=atoi(strtok(&s[7]," \t\n\r"));
    else if (strncasecmp(s,"x-logid:",8)==0)
      logID=atoi(strtok(&s[8]," \t\n\r"));
    else if (strncasecmp(s,"x-compress:",11)==0) {
      docompress=atoi(strtok(&s[11]," \t\n\r"));
      docompress = 0;		/* Disable compression for SMTP */
    } else if (strncasecmp(s,"x-source:",9)==0)
      source=strdup(strtok(&s[9]," \t\n\r"));
    else if (strncasecmp(s,"x-sourcetitle: ",15)==0)
      sourcetitle=strdup(strtok(&s[15],"\n\r"));
    else if (strncasecmp(s,"x-path:",7)==0)
      path=strdup(strtok(&s[7]," \t\n\r"));
    else if (strncasecmp(s,"x-form:",7)==0) {
      char* t=strtok(&s[7]," \t\n\r");
      form=(t[0]=='y' || t[0]=='Y') ? 1 : 0;
    } else if (strncasecmp(s,"x-user:",7)==0)
      user=strdup(strtok(&s[7]," \t\n\r"));
    else if (strncasecmp(s,"x-reply:",8)==0)
      reply=strdup(strtok(&s[8]," \t\n\r"));
    else if (strncasecmp(s,"x-time1:",8)==0) {
      unsigned long ultime;
      sscanf(&s[8],"%lu",&ultime);
      time1=(time_t) ultime;
    } else if (strncasecmp(s,"x-prefetch:",11)==0)
      sscanf(&s[11],"%d",&prefetch);
    else if (strncasecmp(s,"x-preftime:",11)==0)
      sscanf(&s[11],"%ld",&preftime);
    else if (strncasecmp(s,"x-inline:",9)==0) {
      char* t=strtok(&s[9]," \t\n\r");
      doinline=(t[0]=='y' || t[0]=='Y') ? 1 : 0;
    }
  }
  
  if ((long) time2 - (long) time1 < preftime) prefetch=0;
  
  for (;;) {
    if (!fgets(s,1000,stdin)) puke("No body.");	/* get first line of HTTP */
    if (strspn(s," \t\n\r")!=strlen(s)) break;
  }

/* parse the first line */
  method=strtok(s," \t\n\r");
  if (method==NULL) puke("No method.");
  url=strtok(NULL," \t\n\r");
  if (url==NULL) puke("No URL.");
  if (reply==NULL) puke("No reply address.");

/* Figure out how to send mail and call line-mode browser. */
  if (argc>1) {
    mailcmd=(char*) malloc(strlen(argv[1])+strlen(reply)+2);
    sprintf(mailcmd,"%s %s",argv[1],reply);
  }
  else {
    mailcmd=(char*) malloc(strlen(reply)+10);
    sprintf(mailcmd,"/usr/lib/sendmail %s",reply);
  }

  if (argc>2) {
    wwwcmd=(char*) malloc(strlen(argv[2])+strlen(url)+2);
    sprintf(wwwcmd,"%s %s",argv[2],url);
  }
  else {
    wwwcmd=(char*) malloc(strlen(url)+21);
    sprintf(wwwcmd,"www -n -source %s",url);
  }

/*  fprintf(stderr, "Mail command `%s'\n", mailcmd);
  fprintf(stderr, "WWW command `%s'\n", wwwcmd);
 */

/* Get the document. */
  if ((es=fetch(url,method,0,source,sourcetitle,path,form,time1,
	       time2,stdin,prefetch,doinline,NULL)) != NULL)
    puke(es);

  return 0;
}
