 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rover-alone.c
 *
 * Description: Main Rover Server standalone code.
 *
 *
 * Global Functions: None
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

/* #define EXEC */
/* #define TIMING */
/* #define TIMER */

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>

#include "rover.h"
#include "rovertcp.h"
#include "database.h"
#include "RoverZlib.h"

#define TIMEOUT_SECS	1000L
#define TIMEOUT_USECS	500000L

extern char *ServerName;
extern int NoAuth, pending;

/* Library include files */
#include "tcp.h"
#include "HTUtils.h"
#include "HTAccess.h"
#include "HTError.h"
#include "HTThread.h"
#include "HTList.h"
#include "RoverTCP.h"
#include "RoverEvent.h"
#include "html.h"

extern char *stableTraceCallback(ClientData clientData, Tcl_Interp *interp,
				 char *name1, char *name2, int flags);

PRIVATE void RoverSignalsInit(int alarm);
PRIVATE void serverfork();
PRIVATE void servermain();
PRIVATE HTEventState RoverTimer(HTRequest ** request);
PRIVATE HTEventState RoverPortCallback(HTRequest **actreq);
PRIVATE void executeRequest();
PRIVATE void ProcessRequest(ServerState *s);
PRIVATE HTEventState RoverMainLoop(int fd, ServerState *s);
PRIVATE void RoverGetRootInfo();

PRIVATE char line[2050], command[256], output[2048];
PRIVATE HTEventCallBack *RoverCallbackReg;	/* To reg. socket for accept */
PRIVATE int ServerMaxRestarts = 20;
PRIVATE int ServerNoFork = 0;
PRIVATE int ServerSingleExec = 0;
PRIVATE int forked = 0;
PRIVATE int ServerSockFD = -1;
PRIVATE int ServerPort = 9090;
PRIVATE int ServerChild = 0;
PRIVATE int ServerGChild = 0;
PRIVATE int ServerPid = 0;
PRIVATE char logtmp[256];
PRIVATE char *signals[] = {
  NULL,
  "SIGHUP",
  "SIGINT",
  "SIGQUIT",
  "SIGILL",
  "SIGTRAP",
  "SIGSEGV",
  "SIGBUS",
  "SIGFPE",
  "SIGKILL",
  "SIGUSR1",
  "SIGSEGV",
  "SIGUSR2",
  "SIGPIPE",
  "SIGALRM",
  "SIGTERM",
  "SIGSTKFLT",
  "SIGCHLD",
  "SIGCONT",
  "SIGSTOP",
  "SIGTSTP",
  "SIGTTIN",
  "SIGTTOU",
  "SIGURG",
  "SIGXCPU",
  "SIGXFSZ",
  "SIGVTALRM",
  "SIGPROF",
  "SIGWINCH",
  "SIGIO",
  "SIGPWR",
  "SIGUNUSED"
};


#include	<stdio.h>
#include	<signal.h>
#include	<sys/param.h>
#include	<errno.h>
#include 	<sys/resource.h>
extern int	errno;
#ifdef decstation
extern int waitpid(int pid, union wait *status, int options);
#else
extern pid_t waitpid(int , int *, int);
#endif /* decstation */

#ifdef	SIGTSTP		/* true if BSD system */
#include	<sys/file.h>
#include	<sys/ioctl.h>
#endif

#include 	<ctype.h>

#ifdef linux
#define SIG_HANDLER(NAME)	\
PRIVATE void NAME(int sig)
#else
#define SIG_HANDLER(NAME)	\
PRIVATE void NAME(int sig,int code,struct sigcontext *scp,char *addr)
#endif


#ifdef ROVER_TIMERS
/* Set the timer */
PRIVATE void RoverSetTimer()
{
  struct itimerval value, ovalue;

  value.it_value.tv_sec = value.it_interval.tv_sec = TIMEOUT_SECS;
  value.it_value.tv_usec =value.it_interval.tv_usec = TIMEOUT_USECS;
  setitimer(ITIMER_REAL,  &value, &ovalue);
}  

/* Reset the timer */
PRIVATE void RoverResetTimer()
{
  struct itimerval value, ovalue;

  value.it_value.tv_sec = value.it_interval.tv_sec = 
    ovalue.it_value.tv_sec = ovalue.it_interval.tv_sec = 0L;
  value.it_value.tv_usec = value.it_interval.tv_usec =
    ovalue.it_value.tv_usec = ovalue.it_interval.tv_usec = 0L;
  setitimer(ITIMER_REAL,  &value, &ovalue);
}
#endif /* ROVER_TIMERS */

/*
 * Called when we have BIG problems.
 */
SIG_HANDLER(BusSigHandler)
{
  fprintf(stderr, "Received SIGBUS. [BusSigHandler:rover-alone.c]\n");
  roverCloseLogs();
  abort();
}


/*
 * Called when we have BIG problems.
 */
SIG_HANDLER(SegVSigHandler)
{
  fprintf(stderr, "Received SIGSEGV. [SegvSigHandler:rover-alone.c]\n");
  roverCloseLogs();
  abort();
}


/* Called when we get a funny signal. */
SIG_HANDLER(BaseSigHandler)
{
  fprintf(stderr,
	  "WARNING! Ignoring received signal %s (%d). [BaseSigHandler:rover-alone.c]\n",
	  signals[sig], sig);
  /* Catch other bad signals that we ignore (except SIGABRT) */
  signal(SIGQUIT, BaseSigHandler);
  signal(SIGILL, BaseSigHandler);
  signal(SIGTRAP, BaseSigHandler);
  signal(SIGFPE, BaseSigHandler);
#ifdef sparc
  signal(SIGEMT, BaseSigHandler);
  signal(SIGSYS, BaseSigHandler);
  signal(SIGLOST, BaseSigHandler);
#endif
  longjmp(state->RoverEnv, ROVER_SIGNAL);
}


/*
 * Called when a write to a pipe fails.
 */
SIG_HANDLER(PipeSigHandler)
{
  fprintf(stderr, "Received SIGPIPE. [PipeSigHandler:rover-alone.c]\n");
  signal(SIGPIPE, PipeSigHandler);	/* Catch broken pipe signals */
  return;
  /*  longjmp(state->RoverEnv, ROVER_PIPE); */
}

/*
 * Called when a write to a pipe fails.
 */
SIG_HANDLER(AlarmSigHandler)
{
  int i, rc1, rc2;
  struct stat sbuf;

  fprintf(stderr, "Received SIGALRM. [AlarmSigHandler:rover-alone.c]\n");
  signal(SIGALRM, AlarmSigHandler);	/* Catch alarm signals */
#ifdef ROVER_TIMERS
  RoverResetTimer();
#endif /* ROVER_TIMERS */
  if (ServerGChild) {
    fprintf(stderr, "Server GCHILD received bogus SIGALARM. Ignoring...\n");
    return;
  } else if (!ServerChild) {
    fprintf(stderr, "Server PARENT received bogus SIGALARM. Ignoring...\n");
    return;
  }
  fprintf(stderr, "Forking off long running request...");
  reset_auth();		/* Bad, become root so we can open a new log file :( */
  state->pid = fork();
  if (state->pid == -1) {
    perror("Fork failed");
    exit(-1);
  } else if (state->pid == 0) {
    ServerGChild = 1;
    /* Close extra sockets */
    close(ServerSockFD);
    fclose(ServerLogFile);
    fclose(ServerLogFile2);
    for (i = 0; i < MAXSTATES; i++)
      if (States[i].inuse && (state != &(States[i]))
	  && (States[i].reqType != NOTYPE)) {
	if (States[i].Stdin) {
	  fclose(States[i].Stdin);
	  fclose(States[i].Stdout);
	  States[i].Stdin = States[i].Stdout = NULL;
	}
	if (States[i].Query) {
	  free(States[i].Query);
	  States[i].Query = NULL;
	}
	if (States[i].ContentType) {
	  free(States[i].ContentType);
	  States[i].ContentType = NULL;
	}
      }
    RoverSignalsInit(0);

    /*
     * Reinitialize the logs
     */
    sprintf(logtmp, "%s/rover-log-file__%s__%s__%d", ServerLogDir,
	    state->rhost,state->ClientUser,state->msgID);
    if (!stat(logtmp, &sbuf)) {
      /* Opps, file already exists? */
      fprintf(stderr,">>> WARNING! File %s already exists, deleting\n",logtmp);
      unlink(logtmp);
    }
    /* Create a new file */
    rc1 = roverInitLogFile(logtmp, &ServerLogFile, &ServerLogFD);

    sprintf(logtmp, "%s/rover-log-output-file__%s__%s__%d", ServerLogDir,
	    state->rhost,state->ClientUser,state->msgID);
    if (!stat(logtmp, &sbuf)) {
      /* Opps, file already exists? */
      fprintf(stderr,">>> WARNING! File %s already exists, deleting\n",logtmp);
      unlink(logtmp);
    }
    /* Create a new file */
    rc2 = roverInitLogFile(logtmp, &ServerLogFile2, &ServerLogFD2);

    perform_auth_check(state);	/* Take us back, quickly! */
    RoverSetAuth(state);
    if (rc1 || rc2) {
      sprintf(line, "fopen %s failed: %s [AlarmSigHandler:rover-alone.c]\n",
	      logtmp, strerror(errno));
      errorExit(line);
    }
    return;
  } else {
    perform_auth_check(state);	/* Take us back, quickly! */
    RoverSetAuth(state);
    forked = 1;
    fprintf(stderr, "pid is %d\n", state->pid);
    RoverSignalsInit(1);
    longjmp(state->RoverEnv, ROVER_FORK);
  }
}

PRIVATE void RoverSignalsInit(int alarm)
{   
  signal(SIGPIPE, PipeSigHandler);	/* Catch broken pipe signals */
  if (alarm) signal(SIGALRM, AlarmSigHandler);	/* Catch alarm signals */

  /* Catch signals, so we can clean up */
  signal(SIGBUS, BusSigHandler);
  signal(SIGSEGV, SegVSigHandler);
  
  /* Catch other bad signals that we ignore (except SIGABRT) */
  signal(SIGQUIT, BaseSigHandler);
  signal(SIGILL, BaseSigHandler);
  signal(SIGTRAP, BaseSigHandler);
  signal(SIGFPE, BaseSigHandler);
#ifdef sparc
  signal(SIGEMT, BaseSigHandler);
  signal(SIGSYS, BaseSigHandler);
  signal(SIGLOST, BaseSigHandler);
#endif
}

/*
 * Wait on the child standalone server process. If the child ends, return.
 */
PRIVATE void waitchild()
{
#ifdef decstation
  union wait status;
#else
  int status;
#endif
  int pid;
  static int count = 0;

 again:
  pid = waitpid(-1, &status, 0);  /* Get child's status */
  fprintf(stderr, "PARENT: Server's status is %d\n", pid);
  if ((pid == -1) || (pid == ServerPid) ) {	/* No one's alive */
    fprintf(stderr, "PARENT: Server ended unexpectedly...");
  } else {				/* All OK */ 
    fprintf(stderr,
	    "PARENT: Unexpected return! waitpid = %d [waitchild:rover-alone.c]\n",
	    pid);
    goto again;
  }
  if (WIFEXITED(status))
    fprintf(stderr, "normal exit: %d...", WEXITSTATUS(status));
  else if (WIFSIGNALED(status))
    fprintf(stderr, "killed by signal %d (%s)...", WTERMSIG(status),
	    signals[WTERMSIG(status)]);
  count++;
  if (count == ServerMaxRestarts) {
    fprintf(stderr, "PARENT: Max restarts (%d) reached, exiting\n", count);
    exit(0);
  }
}

/*
 * Wait on any grandchildren standalone server processes, if any.
 *   Returns:	0 	if no process has changed status
 *		pid 	if the process ended normally and with a good rc
 *		-pid 	if the process ended abnormally or with a bad rc
 */
PRIVATE int waitgchild()
{
#ifdef decstation
  union wait status;
#else
  int status;
#endif
  int pid;

  pid = waitpid(-1, &status, WNOHANG);  /* Get grandchildren's status */
  if (pid == -1) {
    if (forked) {
      fprintf(stderr,
	      ">>>Forked set, but no children found! [waitgchild:rover-alone.c]\n");
      forked = 0;
    }
    return(0);
  } else if (pid == 0) {
    /* No one ready */
    return(0);
  } else {
    /* Process the child that ended */
    fprintf(stderr, "Grandchild %d ended...", pid);
    if (WIFEXITED(status)) {
      fprintf(stderr, "normal exit: %d\n", WEXITSTATUS(status));
      if (WEXITSTATUS(status) == 0)
	return(pid);
      else
	return(-pid);
    } else if (WIFSIGNALED(status)) {
      fprintf(stderr, "killed by signal %d (%s)\n", WTERMSIG(status),
	      signals[WTERMSIG(status)]);
      return(-pid);
    } else {
      fprintf(stderr, "no information?\n");
      return(-pid);
    }
  }
}

/* 
 * Fork off a child server process. The child sets the ServerChild variable
 *   so that the code and distinguish between the child and server instances.
 */
PRIVATE void serverfork()
{
  ServerPid = fork();
  if (ServerPid == -1) {
    perror(ServerName);
    exit(-1);
  }
  if (ServerPid==0) {
    ServerChild = 1;
  }
}

/* Tell the user how to run the program. */
PRIVATE void usage()
{
  fprintf(stderr,
	  "Usage: %s [-p #] [-F dir] [-H] [-I] [-N] [-V] [-d #] [-f] [-n] [-t] [-u] [-1]\n",
	  ServerName);

  fprintf(stderr, "
-p port\t\t Specify the port the server listens on for requests.
-F dir\t\t Specify the server's logging directory.
-H\t\t Disable the HTTP proxy's hash table.
-I\t\t Use the HTTP proxy's hash table, but ignore the contents.
-N\t\t Disable forking of a separate server process.
-V\t\t Disable tracing and logging of stable variables.
-d level\t Specify the server's debugging level.
-f\t\t Enable synchronous flushing of server log.
-n\t\t Disable all server logging (including stable variables).
-t\t\t Null trace stable variables (No logging or actions done).
-u\t\t Disable user authentication (perform all actions as root).
-1\t\t Disable restart of forked server after a failure.
-R #\t\tLimit the number of server restarts.\n");
  exit(-1);
}

/* Display the server options */
PUBLIC void serverShowOptions()
{
  fprintf(stderr, "Server options:\n");
  fprintf(stderr,
	  "debug level: %d, port: %d, fork: %s, restart: %s (%d), user auth: %s\n",
	  RoverDebug, ServerPort, (ServerNoFork ? "disabled" : "enabled"),
	  (ServerSingleExec ? "disabled" : "enabled"), ServerMaxRestarts,
	  (NoAuth ? "disabled" : "enabled"));
  fprintf(stderr, "Server logging %s",
	  (ServerNoLog ? "disabled\n" : "enabled: "));
  if (!ServerNoLog)
    fprintf(stderr, "sync flush %s, log dir %s\n",
	    (!ServerHardFlush ? "disabled" : "enabled"), ServerLogDir);
  fprintf(stderr, "Stable vars: %s, ",
	  (NullTrace ? "disabled" : 
	   (!ServerUseStableVars ? "ignored" : 
	    (ServerNoLog ? "enabled [but not logged]" : "enabled"))));
  fprintf(stderr, "HTTP proxy hash table: %s\n",
	  (!ServerUseHash ? "disabled" : 
	   (ServerIgnoreHash ? "ignored" : 
	    (ServerNoLog ? "enabled [but not logged]" : "enabled"))));
}

/* This procedure never returns. */
PUBLIC void main(int argc, char **argv)
{
  int a;

  ServerName = argv[0];
  state = &(States[0]);
  
  if (strstr(ServerName, "rmgate")) exit(processHTTP(NULL, argc, argv, 1));

  /* Process command line flags. */
  for (a = 1; a < argc; a++) {
    if (!strcmp(argv[a],"-p")) {		/* port number to listen on */
      a++;
      if (argc > a) ServerPort = atoi(argv[a]);
      else usage();
    } else if (!strcmp(argv[a],"-R")) {		/* Number of server restarts */
      a++;
      if (argc > a) ServerMaxRestarts = atoi(argv[a]);
      else usage();
    } else if (!strcmp(argv[a],"-d")) {		/* server debug level */
      a++;
      if (argc > a) RoverDebug = atoi(argv[a]);
      else usage();
    } else if (!strcmp(argv[a],"-N")) {		/* Don't fork server */
      ServerNoFork = 1;
    } else if (!strcmp(argv[a],"-I")) {	/* Use, but ignore HTTP proxy hsh tbl*/
      ServerIgnoreHash = 1;
    } else if (!strcmp(argv[a],"-V")) {		/* No stable variables */
      ServerUseStableVars = 0;
    } else if (!strcmp(argv[a],"-H")) {		/* No HTTP proxy hash table */
      ServerUseHash = 0;
    } else if (!strcmp(argv[a],"-n")) {		/* No server logging */
      ServerNoLog = 1;
    } else if (!strcmp(argv[a],"-1")) {		/* No server restart */
      ServerSingleExec = 1;
    } else if (!strcmp(argv[a],"-F")) {		/* Server log directory */
      a++;
      if (argc > a) ServerLogDir = argv[a];
      else usage();
    } else if (!strcmp(argv[a],"-f")) {		/* Hard flush log file */
      ServerHardFlush = 1;
    } else if (!strcmp(argv[a],"-t")) {		/* Null variable tracing */
      NullTrace = 1;
    } else if (!strcmp(argv[a],"-u")) {		/* No auth changes */
      NoAuth = 1;
    } else usage();
  }
  gettimeofday(&StartTv, &Tz);
  fprintf(stderr, "Standalone daemon PARENT Rover server started %d:%s",
	  getpid(), ctime((const time_t *)&(StartTv.tv_sec)));

  RoverRootDirsInit();

  if (!ServerLogDir) {
    ServerLogDir = (char *)malloc(strlen(ROVER_SERVER_LOG_DIR)+
				  strlen(RoverRoot[RoverRootID])+2);
    if (!ServerLogDir) {
      fprintf(stderr, "Out of memory [serverinit:rover-alone.c]\n");
      exit(-1);
    }
    sprintf(ServerLogDir, "%s%s",RoverRoot[RoverRootID],ROVER_SERVER_LOG_DIR);
  }
  serverShowOptions();
  if (ServerNoFork) {
    servermain();
    return;
  }
  while (1) {
    serverfork();
    if (ServerChild) {
      servermain();	/* This is the child process. */
    } else {
      waitchild();	/* Wait on the child */
      if (ServerSingleExec) exit(0);
      fprintf(stderr, "restarting\n");
    }
  }
  /* This line should never be reached */
}

PRIVATE void serverinit()
{
  int rc;
  struct stat sbuf;

  gettimeofday(&StartTv, &Tz);
  fprintf(stderr,
	  "Standalone daemon CHILD Rover server %d starting on port %d:%s",
	  getpid(), ServerPort, ctime((const time_t *)&(StartTv.tv_sec)));
  standalone = 1;

#ifdef TIMING
  nice(-40);
  setpriority(PRIO_PROCESS, 0, -20);
#endif /* TIMING */

  RoverGetRootInfo();			/* Save root info */
  HTThreadInit();		        /* Initialize bit arrays */
  outstrInit(1024*1024);		/* Allocate large persistent buffer */
  RoverSignalsInit(1);
  Tcl_InitHashTable(&Info, TCL_STRING_KEYS);	/* User Info htable */

  fclose(stdin);	/* Prevent apps from causing problems */
  fclose(stdout);
  
  /* Open a TCP socket (an Internet stream socket) */
  if ((ServerSockFD = RoverTCPOpenSocket(ServerPort, 0)) < 0) {
    perror("server: can't open server socket");
    exit(1);
  }

  if (RoverDebug) fprintf(stderr, "Server socket is %d\n", ServerSockFD);
  RoverCallbackReg = calloc(sizeof(HTEventCallBack),1);
  RoverCallbackReg->sockfd = ServerSockFD;
  RoverCallbackReg->callback = RoverPortCallback;
  HTEventRegister(RoverCallbackReg, EVENT_READ);	  /* Register client */

  /*
   * Initialize the log
   */
  if (stat(ServerLogDir, &sbuf)) {
    fprintf(stderr, "Creating Rover Server log directory (%s)\n",ServerLogDir);
    if ((rc = mkdir(ServerLogDir, 0777)) != 0) {
      fprintf(stderr,
	      "Error creating server log dir (%s) %d [serverinit:rover-alone.c]\n",
	      command, rc);
      exit(-1);
    }
  } else if (!S_ISDIR(sbuf.st_mode)) {
    fprintf(stderr,
	    "Log dir exists, but not as dir [serverinit:rover-alone.c]\n");
    exit(-1);
  }

  sprintf(command, "%s/rover-log-file", ServerLogDir);
  if (!stat(command, &sbuf)) recover(command, NULL);

  sprintf(command, "%s/rover-log-output-file", ServerLogDir);
  if (!stat(command, &sbuf)) recover(command, NULL);

  /* Append to the log files */
  sprintf(command, "%s/rover-log-file", ServerLogDir);
  rc = roverInitLogFile(command, &ServerLogFile, &ServerLogFD);
  if (rc) {
    fprintf(stderr, "Open of file %s failed [serverinit:rover-alone.c]\n",
	    command);
    exit(-1);
  }
  sprintf(command, "%s/rover-log-output-file", ServerLogDir);
  rc = roverInitLogFile(command, &ServerLogFile2, &ServerLogFD2);
  if (rc) {
    fprintf(stderr, "Open of file %s failed [serverinit:rover-alone.c]\n",
	    command);
    exit(-1);
  }

  HTTimerRegister(RoverTimer, 3);
}

PRIVATE HTEventState RoverTimer(HTRequest ** request)
{
  HTEventState status = EVENT_OK;
  if (RoverDebug) fprintf(stderr, "Timer callback\n");
  if (!pending) cleanLog();
  return(status);
}


/* Returns -1 if no children to rerun, otherwise returns slot i to rerun */
PRIVATE int checkGChildren()
{
  int j, i, pid = waitgchild();

  if (pid) {
    /* A long running child ended, check if it needs to be cleaned up
     *   or re-run
     */
    if (pid < 0) {
      /* Need to re-run it */
      fprintf(stderr, ">>>> Need to rerun long running child <<<<\n");
      for (pid = -pid, j = -1, forked = i = 0; i < MAXSTATES; i++)
	if (States[i].inuse) {
	  if (States[i].pid == pid) {
	    j = i;
	    States[i].pid = 0;
	  } else if (States[i].pid)
	    forked = 1;
	}
      if (j == -1)
	fprintf(stderr, ">>>> Request for pid %d not found!\n", pid);
      else 
	return(j);
    } else {
      /* Need to clean up after it */
      fprintf(stderr, "Cleaning up after long running child\n");
      for (forked = i = 0; i < MAXSTATES; i++)
	if (States[i].inuse) {
	  if (States[i].pid == pid) {
	    reset_state(&(States[i]));
	  } else if (States[i].pid)
	    forked = 1;
	}
    }
  }
  return(-1);
}

PRIVATE ServerState *pendingCheck()
{
  int rc = 0, i, found = 0;

  if (forked & ((i = checkGChildren()) != -1)) {
    return(&(States[i]));		/* Retry child */
  } else if (ServerRecovering == NORMALOP) {
    for (found = i = 0; i < MAXSTATES; i++) {
      if (States[i].inuse && !States[i].pid) {
	rc = 0;
	if (States[i].reqType != NOTYPE) {
	  if ((rc = checkMsgID(States[i].ClientUser, States[i].rhost,
			       States[i].msgID, States[i].incarnation,
			       States[i].retry)) == 1) {
	    fprintf(stderr, ">>>Msg from %s@%s: %d now acceptable\n",
		    States[i].ClientUser, States[i].rhost,States[i].msgID);
	    return(&(States[i]));
	  }
	}
	if (rc == -1) {
	  if (States[i].Stdin) {
	    if (!States[i].keepalive && (States[i].tcp->sock != -1)) {
	      close(States[i].tcp->sock);
	      States[i].tcp->sock = -1;
	      fclose(States[i].Stdin);
	      fclose(States[i].Stdout);
	    }
	    States[i].Stdin = States[i].Stdout = NULL;
	  }
	  reset_state(&(States[i]));
	  continue;
	} else {
	  if (!pending) fprintf(stderr, "Opps got !pending\n");
	  pending = found = 1;
	}
      }
    }
    if (!found) pending = 0;
  }
  return(NULL);
}

PRIVATE void servermain()
{
  int rc;  
  HTEventState status = EVENT_OK;

  serverinit();
  while (ServerRecovering != NORMALOP) {
    /* Find the next message to deal with during recovery */
    int i, j, min = MAXSTATES;
    for (j = -1, i = 0; i < MAXSTATES; i++) 
      if (States[i].inuse && (States[i].order < min)
	  && (States[i].reqType != NOTYPE)) {
	min = States[i].order;
	j = i;
      }
    if (min == MAXSTATES) {
      ServerRecovering = NORMALOP;
      fprintf(stderr, "Server done with log recovery\n");
      continue;
    } else {
      state = &(States[j]);
      state->openConnection = 1;
    }
    state->state = TCP_EXECUTE;
    status = RoverMainLoop(state->tcp->sock, state);
/*      executeRequest(); */
  }

  /* Begin normal operation */
  if ((rc = HTEventLoopNull()) == HT_ERROR) {
    fprintf(stderr, "Rover Server unable to start eventloop\n"); 
    exit(-1);
  }
}

PRIVATE void RoverTCPReset(ServerState *s, TCP_RESET type)
{
  reset_state(s);

#if 0

  if (type == ABORT) {
    switch (http->reqType) {
    case CACHE:
      /* >>> Delete cache entry and file <<< */
      break;
    case OBJECT:
      if (http->stream)
	(*http->stream->isa->abort)(http->stream, NULL);
      break;
    case CLIENT:
    case NOTYPE:
      /* Nothing to do */
      break;
    default:
      fprintf(stderr, "Got unknown HTTP state type during abort\n");
      http->reqType = NOTYPE;
      break;
    }
  }

  if ((type == FULL) || (type == ABORT)) {
    HTTPStates[http->tcp->sock] = NULL;
    free(http);
  }
  
  if (http->inflate) {
    EndUncompress(http->inflate);
    http->inflate = NULL;
  }
#endif
}

PRIVATE void TCPProcessMIME(ServerState *s)
{
  int size = (s->compress) ? s->uncompressLen+1 : s->ContentLength+1;

  initState(s);
  s->first = 1;

  if (!s->ContentType) {
    s->status = 400;
    errorExit("Server called without a CONTENT_TYPE");
    s->state = TCP_ABORT;
    return;
  }
  if (strcmp(s->ContentType, "application/x-rover")) {
    sprintf(line, "Got CONTENT_TYPE: %s; not application/x-rover",
	    s->ContentType);
    s->status = 400;
    errorExit(line);
    s->state = TCP_ABORT;
    return;
  }

  s->keepalive = 0;
  if ((s->Query = malloc(size)) == NULL) {
    errorExit("Server out of memory");
    s->state = TCP_ABORT;
    return;
  }
   
  if (s->compress) {
    s->bufferOffset = 0;
    if (!(s->inflate = InitUncompress())) { 
      fprintf(stderr,"InitUncompress err [ProcessMIMELine:rover-alone.c]\n");
      s->state = TCP_ABORT;
      return;
    }
  }
  if (s->tcp->size) {
    if (RoverDebug) fprintf(stderr, "Saving %d bytes from buffer\n",
			      s->tcp->size);
    s->tcp->offset = 0;
    s->state = TCP_SAVE_BODY;	/* Save the existing body */
    return;
  }
  s->bufferOffset = s->tcp->size = s->tcp->offset = 0;
  s->state = TCP_READ_BODY;	/* Read the body */
}


/* Each line starts with the name of a field, followed by its value.
   Booleans are "yes" or "no"; others types are formatted as expected.
   When it gets to the "url" field, it starts a new record: either
   reads in new values for one that's there, or creates a new one if
   none match.

   Returns -1 if there is an error. Returns 0 if more lines are expected;
   otherwise, returns 1 if no more lines are expected.
 */
PRIVATE int ProcessMIMELine(void *ClientData, char *ptr, void *nextState)
{
  char *t;

  state = (ServerState *) ClientData;
  if (ptr == NULL) {
    if ((int) nextState == -1) 
      state->state = TCP_ABORT;
    else
      state->state = (TCPState) nextState;
    return(1);
  }
    
  if ((*ptr == '\n')||((*ptr == '\r') && (*(ptr+1)=='\n'))){/* End of header */
    if (RoverDebug)
      fprintf(stderr,"Done processing MIME hdr lines\n");
    return(1);				/* Go to TCP_READ_BODY state */
  }
  
  if ((t = strchr(ptr, '\n')) != NULL) *t = 0;		/* Delete \n */
  if (!strncasecmp(ptr, "content-type:", 13)) {		/* doc type */
    state->ContentType = strdup(strtok(&ptr[13], " \t\n\r"));
/*    fprintf(stderr, "Got ContentType %s\n", state->ContentType); */
  } else if (!strncasecmp(ptr, "content-length:", 15)) { 	/* doc len */
    state->ContentLength = atoi(strtok(&ptr[15], " \t\n\r"));
/*    fprintf(stderr, "Got ContentLength %d\n", atoi(value)); */
  } else if (!strncasecmp(ptr, "compress:", 9)) { 	/* doc compressed? */
    state->compress = atoi(strtok(&ptr[9], " \t\n\r"));
/*    fprintf(stderr, "Got CompressLevel %d\n", atoi(value)); */
  } else if (!strncasecmp(ptr, "length:",7)) { 		/* doc raw len */
    state->uncompressLen = atoi(strtok(&ptr[7], " \t\n\r"));
/*    fprintf(stderr, "Got UncompressLength %d\n", atoi(value)); */
  }
  return(0);
}

PUBLIC char *TcpText(TCPState state)
{
  switch (state) {
  case TCP_BEGIN:
    return("TCP_BEGIN");
    break;
  case TCP_NEED_REQUEST:
    return("TCP_NEED_REQUEST");
    break;
  case TCP_GOT_REQUEST:
    return("TCP_GOT_REQUEST");
    break;
  case TCP_READ_MIMEHDR:
    return("TCP_READ_MIMEHDR");
    break;
  case TCP_PROC_MIMEHDR:
    return("TCP_PROC_MIMEHDR");
    break;
  case TCP_PREP_REQUEST:
    return("TCP_PREP_REQUEST");
    break;
  case TCP_READ_BODY:
    return("TCP_READ_BODY");
    break;
  case TCP_SAVE_BODY:
    return("TCP_SAVE_BODY");
    break;
  case TCP_GOT_BODY:
    return("TCP_GOT_BODY");
    break;
  case TCP_WAIT_MSGID:
    return("TCP_WAIT_MSGID");
    break;
  case TCP_EXECUTE:
    return("TCP_EXECUTE");
    break;
  case TCP_CLEANUP:
    return("TCP_CLEANUP");
    break;
  case TCP_ABORT:
    return("TCP_ABORT");
    break;
  default:
    return("<unknown>");
    break;
  }
}

/* Change a request's state */
PRIVATE void updateState(void *ClientData, void *newState)
{
  ((ServerState *)ClientData)->state = ((TCPState)newState);
}

/* Wrapper for callbacks on sockets */
PRIVATE HTEventState RoverTCPCallback(HTRequest **actreq)
{
  int sockReady = 1;  
  HTEventState status = EVENT_OK;

  state = NULL;
  while (sockReady || ((state = pendingCheck()) != NULL)) {
    /* Before receiving any new messages:
       1. Check if any long running children have exited
       2. If there are any pending messages that are now acceptable.
       */
    if (RoverDebug) fprintf(stderr, "sockReady is %d, state is 0x%x [2]\n",
			    sockReady, (unsigned int) state);

    if (state || (state = pendingCheck())) {
      state->state = TCP_EXECUTE;
      status = RoverMainLoop(state->tcp->sock, state);
/*      executeRequest(); */
      continue;
    }

    sockReady = 0;		/* Don't try again */
    status = RoverMainLoop((int) actreq, TCPStates[(int)actreq]);
  }
  if (RoverDebug) fprintf(stderr, "End of [RoverTCPCallback:rover-alone.c]\n");
  return(status);
}

PRIVATE HTEventState RoverMainLoop(int fd, ServerState *s)
{
  HTEventState status = EVENT_OK;
  int count, rc;
  char tmp;

  state = s;
  /* Now jump into the machine. We know the state from the previous run */
  while (1) {
    if (RoverDebug)
      fprintf(stderr,
	      "TCP state for %d is %s [RoverTCPCallback:rover-alone.c]\n",
	      (int) fd, TcpText(state->state));
    switch (state->state) {
    case TCP_BEGIN:
      fprintf(stderr, "Ended up in the TCP_BEGIN state!\n");
      state->state = TCP_CLEANUP;
      break;
	    
    case TCP_NEED_REQUEST:	  /* Get the client's request */
      if (!TCPGetLine(state->tcp, updateState, (void *) state,
		      (void *) TCP_GOT_REQUEST, (void *) TCP_ABORT))
	return(status);
      break;
    
    case TCP_GOT_REQUEST:
      if (RoverDebug) fprintf(stderr, "Request line: %s", state->tcp->line);
      if (strncmp(state->tcp->line,"POST ", 5)) {	/* Check the method */
	fprintf(stderr, "Got invalid method != POST : %s\n", state->tcp->line);
	state->state = TCP_ABORT;
	continue;
      }
      state->tcp->offset = state->tcp->size = 0;
      state->state = TCP_READ_MIMEHDR;
      break;
      
    case TCP_READ_MIMEHDR:
      if (!TCPRead(state->tcp, state->tcp->line, BUFSIZE, 0,
		   updateState, (void *) state,(void *) TCP_PROC_MIMEHDR,
		   (void *) TCP_PREP_REQUEST, (void *) TCP_ABORT))
	return(status);
      break;

    case TCP_PROC_MIMEHDR:
      state->state = TCP_READ_MIMEHDR;
      rc = NLProcessLine((void *)state, ProcessMIMELine, state->tcp->line,
			 state->tcp->size, (void *) TCP_PREP_REQUEST);
      if (rc == -1) continue;
      state->tcp->size -= rc;
      state->tcp->offset = state->tcp->size;
      break;

    case TCP_PREP_REQUEST:
      TCPProcessMIME(state);
      break;

    case TCP_READ_BODY:
      if (RoverDebug)
        fprintf(stderr,
		"(before) state->tcp->size is %d, state->tcp->offset is %d\n",
                state->tcp->size, state->tcp->offset);
      if (!TCPRead(state->tcp, state->tcp->line, BUFSIZE, state->compress,
		   updateState, (void *) state,(void *) TCP_SAVE_BODY,
		   (void *) TCP_GOT_BODY, (void *) TCP_ABORT))
	return(status);
      if (RoverDebug) {
        tmp = state->tcp->line[state->tcp->size];
        state->tcp->line[state->tcp->size] = '\0';
        fprintf(stderr,	"(after) state->tcp->size is %d, state->tcp->offset is %d `%s'\n",
                state->tcp->size, state->tcp->offset, state->tcp->line);
        state->tcp->line[state->tcp->size] = tmp;
      }
      break;

    case TCP_SAVE_BODY:
      rc = count = state->tcp->size - state->tcp->offset;
      if (!state->compress) {
	if (state->count + count > state->ContentLength)
	  count = state->ContentLength - state->count;
	memcpy(state->Query+state->count, state->tcp->line+state->tcp->offset,
	       count);
        state->count += count;
        state->tcp->offset = state->tcp->size = 0;
	if (state->count >= state->ContentLength) {
	  if (RoverDebug && (rc != count))
	    fprintf(stderr, "Got %d leftover data bytes\n", rc-count);
	  state->state = TCP_GOT_BODY;
	} else {
	  state->state = TCP_READ_BODY;
	  return status;
	}
      } else {
        /* For now, ignore leftover (it was already saved)
        state->bufferOffset = 0; */
        if (RoverDebug)
          fprintf(stderr, "Inflating %d input bytes (%d leftover output bytes) [TCP_SAVE_BODY]\n",
                  state->tcp->size, state->bufferOffset);
        count = 0;
	rc = UncompressBuffer(state->inflate, state->tcp->line, 
			      &(state->tcp->size), state->Query,
			      &(state->bufferOffset), state->uncompressLen,
			      &count);
        if (rc == 1) {
          state->state = TCP_GOT_BODY;
        } else if (rc == 2) {
          state->state = TCP_GOT_BODY;
          /* Leftover data, discard it for now */
          if (RoverDebug)
            fprintf(stderr,"Discarding %d bytes from buffer [TCP_SAVE_BODY]\n",
                    state->tcp->size);
        } else if (rc == -1) {
          fprintf(stderr, "Uncompress error\n");
          state->state = TCP_ABORT;
        } else {
	  state->state = TCP_READ_BODY;
	  return status;
	}
      }
      break;

    case TCP_GOT_BODY:
      if (state->compress) {
	if (state->bufferOffset != state->uncompressLen) {
	  fprintf(stderr,"Uncompress error: got %d, but expected %d bytes\n",
		  state->bufferOffset, state->uncompressLen);
	  state->state = TCP_ABORT;
	}
	state->Query[state->uncompressLen] = '\0';
	state->tcp->offset = state->bufferOffset = 0;
	state->inflate = NULL;
      }
      if (RoverDebug) fprintf(stderr, "Done receiving request.\n");
      ProcessRequest(state);
      break;

    case TCP_WAIT_MSGID:
      if (!(rc = checkMsgID(state->ClientUser, state->rhost, state->msgID,
			    state->incarnation, state->retry))) {
	/* Msg not ready for processing */
	pending = 1;
	state->state = TCP_WAIT_MSGID;
	return(status);
      } else if (rc == -1) {
	if (state->Stdin) {
	  if (!state->keepalive && (state->tcp->sock != -1)) {
	    close(state->tcp->sock);
	    state->tcp->sock = -1;
	    fclose(state->Stdin);
	    fclose(state->Stdout);
	  }
	  state->Stdin = state->Stdout = NULL;
	}
	state->state = TCP_ABORT;
	return(status);
      }

    case TCP_EXECUTE:
      /* Message ready for processing */
      executeRequest();
      state->state = TCP_CLEANUP;
      break;

    case TCP_CLEANUP:
      if (RoverDebug) fprintf(stderr, "Done processing request.\n\n");
      RoverTCPReset(state, FULL);
      return(EVENT_OK);
  
    case TCP_ABORT:
      if (RoverDebug) fprintf(stderr, "Aborting request.\n\n");
      RoverTCPReset(state, ABORT);
      return(EVENT_OK);

    default:
      fprintf(stderr, "Got unknown TCP state %d\n", state->state);
      state->state = TCP_ABORT;
      break;
    }
  } /* End of while(1) */
  return status;
}


/* Called when there's a pending TCP connection (or error). */
PRIVATE HTEventState RoverPortCallback(HTRequest **actreq)
{ 
  int sock, i, sockReady = 1;  
  HTEventState status = EVENT_OK;

  state = NULL;
  while (sockReady || ((state = pendingCheck()) != NULL)) {
    /* Before receiving any new messages:
       1. Check if any long running children have exited
       2. If there are any pending messages that are now acceptable.
       */
    if (RoverDebug) fprintf(stderr, "sockReady is %d, state is 0x%x\n",
			    sockReady, (unsigned int) state);

    if (state || (state = pendingCheck())) {
      state->state = TCP_EXECUTE;
      status = RoverMainLoop(state->tcp->sock, state);
/*      executeRequest(); */
      continue;
    }

    if (!pending) cleanLog();
    
    /* First, find an empty state entry */
    for (i = 0; i < MAXSTATES; i++)
      if (!States[i].inuse) break;
    if (i == MAXSTATES) {
      /* We should have a more graceful choice here */
      fprintf(stderr, "Out of message slots\n");
      exit(-1);
    }
    state = &(States[i]);
    state->compress = 0;
    sockReady = 0;		/* Don't try again */
    sock = RoverTCPAccept(ServerSockFD, 1);
    TCPStates[sock] = state;

#ifdef TIMER
    gettimeofday(&ServerStartTv, &Tz);
#endif /* TIMER */
    if (RoverDebug) fprintf(stderr, "Got connection, fd = %d\n", sock);

    RoverGetRemoteInfo(sock, &(state->remoteName), &(state->remoteIP));
    
    /* Only do the setup first time through. This is actually state
       TCP_BEGIN but it can't be in the state machine as we need the
       structure first */
    pending = 1;
    state->inuse = 1;
    state->state = TCP_BEGIN;
    state->reqType = NOTYPE;
    state->tcp = RoverTCPInit(sock, BUFSIZE);
    state->callback = calloc(sizeof(HTEventCallBack),1);
    state->callback->sockfd = state->tcp->sock;
    state->callback->callback = RoverTCPCallback;
    HTEventRegister(state->callback, EVENT_READ);	  /* Register client */
/*    RoverSetup(state); */
    state->state = TCP_NEED_REQUEST;
    status = RoverTCPCallback((HTRequest **)sock);
  }
  return(status);
}


PRIVATE void ProcessRequest(ServerState *s)
{
  int rc;

  if (RoverDebug) fprintf(stderr, "Processing request\n");
  s->reqType = ROVER_REQ;

  /* ######## For now, block on output ######## */
  fcntl(s->tcp->sock, F_SETFL,
	(fcntl(s->tcp->sock, F_GETFL, 0) & ~(O_NDELAY|O_NONBLOCK)));

  /* Process the new request */
  logClean = 0;		/* Mark log as dirty */
  s->Stdin = fdopen(s->tcp->sock, "r");
  s->Stdout = fdopen(s->tcp->sock, "w");
  if ((rc = setjmp(s->RoverEnv)) == 0) {
    /* Normal, direct call */
    process_query(s);
  } else {
    /* Opps, an errorExit occurred */
    if (rc == ROVER_FATAL) {
      fprintf(stderr, "Rover server terminating due to fatal error!\n");
      exit(-1);
    } 
    fprintf(stderr, "Rover server cleaning up after error\n");
    s->state = TCP_CLEANUP;
    return;
  }
  rc = checkMsgID(s->ClientUser, s->rhost, s->msgID, s->incarnation,
		  s->retry);
  if (!rc) {		/* Msg not ready for processing */
    pending = 1;
    s->state = TCP_WAIT_MSGID;
    return;
  } else if (rc == -1) {
    if (s->Stdin) {
      if (!s->keepalive && (s->tcp->sock != -1)) {
	close(s->tcp->sock);
	s->tcp->sock = -1;
	fclose(s->Stdin);
	fclose(s->Stdout);
      }
      s->Stdin = s->Stdout = NULL;
    }
    s->state = TCP_ABORT;
    return;
  }

  /* Message ready for processing */
  s->state = TCP_EXECUTE;
}

PRIVATE void RoverHandleError(int rc, ServerState *s)
{
  switch (rc) {
  case ROVER_FATAL:
    fprintf(stderr, "Rover %sserver terminating due to fatal error!\n",
	    (ServerGChild ? "GCHILD" : ""));
    if (s->currdb.db) {
      db_close(&(s->currdb));
      s->currdb.db = NULL;
      s->db = NULL;
    }
    saveResult(s->ServerSeq, rc);
    roverCloseLogs();
    exit(-1);
    break;
  case ROVER_FORK:
    fprintf(stderr, "Rover %sserver partial clean up after fork\n",
	    (ServerGChild ? "GCHILD" : ""));
    /* #### How to deal with Tcl interpreter, if any? 
       The fork may have occurred during the execution of a Tcl
       command, leaving the interpreter in an undefined state. 
     */
    if (globalTclInterp) deleteTclInterp();
    signal(SIGALRM, AlarmSigHandler);	/* Catch alarm signals */
    if (s->currdb.db) {
      db_close(&(s->currdb));
      s->currdb.db = NULL;
      s->db = NULL;
    }
    break;
  case ROVER_PIPE:
    fprintf(stderr, "Rover %sserver cleaning up after pipe error\n",
	    (ServerGChild ? "GCHILD" : ""));
    break;
  case ROVER_ERROR:
    fprintf(stderr, "Rover %sserver cleaning up after error\n",
	    (ServerGChild ? "GCHILD" : ""));
    if (s->currdb.db) {
      db_close(&(s->currdb));
      s->currdb.db = NULL;
      s->db = NULL;
    }
    saveResult(s->ServerSeq, rc);
    break;
  case ROVER_SIGNAL:
    fprintf(stderr, "Rover %sserver cleaning up after signal\n",
	    (ServerGChild ? "GCHILD" : ""));
    break;
  default:
    fprintf(stderr,
	    "Rover %sserver got unknown longjmp error %d, terminating\n",
	    (ServerGChild ? "GCHILD" : ""), rc);
    saveResult(s->ServerSeq, rc);
    roverCloseLogs();
    exit(-1);
  }
}

PRIVATE void executeRequest()
{  
  int rc;

  /* Log the start of execution of this request */
  if (!state->ServerSeq) {
    fprintf(stderr, "Missing server sequence ID for message\n");
    exit(-1);
  }
  if ((ServerRecovering == NORMALOP) && !ServerNoLog) {
    fprintf(ServerLogFile, "ROVER_LOG_START %d\n", state->ServerSeq);
    fflush(ServerLogFile);
    if (ServerHardFlush) fsync(ServerLogFD);
  }
  
#ifdef ROVER_TIMERS
  RoverSetTimer();
#endif /* ROVER_TIMERS */
  if ((rc = setjmp(state->RoverEnv)) == 0) {
    /* Normal, direct call */
    rovermain(state, 0, NULL);
#ifdef ROVER_TIMERS
    RoverResetTimer();
#endif /* ROVER_TIMERS */
  } else {
    /* Opps, an errorExit occurred */
#ifdef ROVER_TIMERS
    RoverResetTimer();
#endif /* ROVER_TIMERS */
    RoverHandleError(rc, state);
  }

  if (state->Stdin) {
    if (!state->keepalive && (state->tcp->sock != -1)) {
      close(state->tcp->sock);
      state->tcp->sock = -1;
      fclose(state->Stdin);
      fclose(state->Stdout);
    }
    state->Stdin = state->Stdout = NULL;
  }
  if (rc != ROVER_FORK) {
    reset_state(state);
  } else {
    /* Record the fork in the log */
    if (!ServerNoLog) {
      fprintf(ServerLogFile, "ROVER_LOG_FORK %d %d\n", state->ServerSeq,
	      state->pid);
      fflush(ServerLogFile);
      if (ServerHardFlush) fsync(ServerLogFD);
    }
  }
  if (ServerGChild) {
    fprintf(stderr, "Long running request (pid: %d) is done\n", getpid());
    exit(0);
  }
  reset_auth();
  
  /* Reset database information */
  state->currdb.write = 0;
  state->currdb.newdb = 0;
  state->currdb.db = NULL;
  state->currdb.filename = NULL;
}

/*
 * Reset the server's state
 */
PUBLIC void reset_state(ServerState *s)
{
  int i;

  globalTclInterp = NULL;
  if (!s->inuse) return;
  if (s->Query) {
    if (!s->keepalive) free(s->Query);	/* Query needed by next request */
    s->Query = NULL;
    s->QuerySize = s->ServiceID = 0;
    s->httpdHost = s->Service = s->Object = s->command = NULL;
    s->code = s->password = NULL;
  } else {
    /* Several variables were copied (during recovery) */

    if (s->httpdHost) {
      free(s->httpdHost);
      s->httpdHost = NULL;
    }
    if (s->Service) {
      free(s->Service);
      s->Service = NULL;
    }
    if (s->Object) {
      free(s->Object);
      s->Object = NULL;
    }
    if (s->command) {
      free(s->command);
      s->command = NULL;
    }
    if (s->code) {
      free(s->code);
      s->code = NULL;
    }
    if (s->password) {
      free(s->password);
      s->password = NULL;
    }
  }
  if (s->ContentType) {
    free(s->ContentType);
    s->ContentType = NULL;
  }
  if (s->recoverProc) {
    free(s->recoverProc);
    s->recoverProc = NULL;
  }

  if (s->outputBuffer) {
    free(s->outputBuffer);
    s->outputBuffer = NULL;
  }

  if (s->remoteIP) {
    free(s->remoteIP);
    s->remoteIP = NULL;
  }
  if (s->remoteName) {
    free(s->remoteName);
    s->remoteName = NULL;
  }

  s->retry = s->port = s->pid = s->inuse = s->openConnection = s->order = 0;
  s->ServerSeq = s->ContentLength = s->loggedOutput = 0;
  s->rhost = NULL;
  s->ClientUser = "";
  s->compress = s->first = s->openConnection = s->loggedOutput = 0;
  roverResetAuth(s);

  s->count = s->bufferOffset = 0;
  s->Stdin = s->Stdout = NULL;

  /* Reset environmental information */
  if (state->Home) {
    free(state->Home);
    state->Home = NULL;
  }
  if (state->FastHome) {
    free(state->FastHome);
    state->FastHome = NULL;
  }

#ifndef sparc
  unsetenv("HOME");
  unsetenv("USER");
#endif

  for (i = 0; i < MAXVAR; i++)
    if (s->vars[i].inuse) {
      s->vars[i].inuse = 0;
      if (s->vars[i].name) {
	free(s->vars[i].name);
	s->vars[i].name = NULL;
      }
      if (s->vars[i].value) {
	free(s->vars[i].value);
	s->vars[i].value = NULL;
      }
    }

  if (s->output) {
    outstrReleaseBuffer(s->output);
    s->output = NULL;
  }
  s->Header[0] = '\0';

  /* If this is an abort, delete everything. Otherwise, only delete things
     if the keepalive bit is not set */
  if ((s->state == TCP_ABORT) || (s->callback && !s->keepalive)) {
    if (s->tcp) {
      if (s->tcp->sock != -1) close(state->tcp->sock);
      TCPStates[s->tcp->sock] = NULL;
      RoverTCPFree(s->tcp);
    }
    if (s->callback) {
      HTEventDeregister(s->callback, EVENT_READ);
      HTEventDeregister(s->callback, EVENT_WRITE);
      free(s->callback);
    }
  }
  s->nextState = s->state = TCP_BEGIN;
  s->callback = NULL;
  s->reqType = NOTYPE;
  s->tcp = NULL;
  if (s->inflate) {
    EndUncompress(s->inflate);
    s->inflate = NULL;
  }
  s->keepalive = 0;

  if (s->currdb.db) {
    db_close(&(s->currdb));
    s->currdb.db = NULL;
    s->db = NULL;
    s->currdb.write = 0;
    s->currdb.newdb = 0;
    s->currdb.filename = NULL;
  }
}

PRIVATE void RoverGetRootInfo()
{
  struct passwd *entry = NULL;
  int idx = 0;

 retry:
  if ((entry = getpwnam("root")) == NULL) {
    if (idx < 2) {
      fprintf(stderr, "Tried name `root'\n");
      idx++;
      setpwent();
      goto retry;
    }
    fprintf(stderr,
	    "Rover server - Unrecoverable fatal error: Unable to reset authorization to root!");
    exit(-1);
  }

  /* Save the appropriate groups and uid */
  RoverUid = entry->pw_uid;
  RoverGid = entry->pw_gid;

  RoverHomeEnv =(char *)malloc(strlen("HOME=")+1);
  if (!RoverHomeEnv)
    fatalError("Server unable to allocate environment var. buffer");
  sprintf(RoverHomeEnv, "HOME=");

  RoverUserEnv=(char *)malloc(strlen("USER=")+1);
  if (!RoverUserEnv)
    fatalError("Server unable to allocate environment var. buffer");
  sprintf(RoverUserEnv, "USER=");
  putenv(RoverHomeEnv);
  putenv(RoverUserEnv);
}

/*
 * Construct and send a status header to the client, followed
 *   by the output buffer and any dependency information.
 */
PUBLIC int sendServerResponse(int *extra, int noCompress)
{
  int n, rc = 0;
  z_stream *c_stream = NULL;	/* Compression stream */

  if (standalone && (state->loggedOutput == 1)&& (ServerRecovering != READLOG))
    stableLogOutput();

  generateServerHeader(state);

  if (state->openConnection) {
    /* Open a connection to the client here */
    if (!state->httpdHost || !state->port) {
      fprintf(stderr, "Missing httpdHost (%s) or port (%d) information\n",
	      state->httpdHost, state->port);
      return(0);
    }
    fprintf(stderr, "Opening HTTP connection to the client\n");
    if ((state->tcp->sock = openTCP(state->httpdHost, state->port)) == -1) {
      fprintf(stderr,"openTCP to %s:%d failed\n",state->httpdHost,state->port);
      exit(-1);
    }
    state->Stdin = fdopen(state->tcp->sock, "r");/* Convert to streams. */
    state->Stdout = fdopen(state->tcp->sock, "w");
    if (!state->Stdin || !state->Stdout) {
      fprintf(stderr, "fdopen(state->tcp->sock) failed");
      exit(-1);
    }
  }

  /* Output the header */
  fputs(state->Header, state->Stdout);
  fflush(state->Stdout);

  if (!state->OutSize && !state->DVSize) return(rc);

  if (!noCompress && state->compress) {
    if (!(c_stream = InitCompress(state->compress))) {
      fprintf(stderr, "InitCompress error");
      exit(-1);
    }

    n = 0;
    if (Compress(c_stream, state->Stdout, state->output->buffer,state->OutSize,
		 output,2048,&n) != Z_OK) {
      fprintf(stderr, "Compress error");
      exit(-1);
    }
    rc = n;
    if (Compress(c_stream, state->Stdout, state->DVbuf, state->DVSize, output,
		 2048, &n) != Z_OK) {
      fprintf(stderr, "Compress error");
      exit(-1);
    }
    rc += n;
    *extra = 0;

    if (EndCompress(c_stream, state->Stdout, output, 2048, &n) != Z_OK) {
      fprintf(stderr, "EndCompress error");
      exit(-1);
    }
    rc += n;
  } else {
    rc = fwrite(state->output->buffer, sizeof(char), state->OutSize,
		state->Stdout);
    *extra = fwrite(state->DVbuf, sizeof(char), state->DVSize, state->Stdout);
  }

  fflush(state->Stdout);
  return(rc);
}

