 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rover-cgi.c
 *
 * Description: Main Rover Server CGI module.
 *
 *
 * Global Functions: See HTML documentation
 * Global Variables: See HTML documentation
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

/* #define EXEC */
/* #define TIMING */
/* #define TIMER */

#include "rover.h"
#include "rovertcp.h"

extern char *ServerName;

PRIVATE char line[2050], output[2048];

PRIVATE char *getquery(FILE *stream)
{
  int count, i, rc;
  char *buf, *ptr, c = '\0';

  i = count = state->ContentLength;
  if (count <= 0) return(NULL);
  if ((ptr = buf = malloc(count+1)) == NULL) return(NULL);
  while (1) {
    rc = fread(&c, sizeof(char), 1, stream);
    if (rc != 1) {
      c = '\0';
      break;
    }
    if ((c != ' ') && (c != CR) && (c != LF)) break;
  }
/*  fprintf(stderr, "`%c' is %d\n", c, (int)c); */
  if (c != '\0') {
    buf[0] = c;
    ptr = buf+1;
    i--;
  }
  rc = fread(ptr, sizeof(char), i, stream);
  if (rc != i) {
    fprintf(stderr, "getquery got short read (%d)\n", rc);
    free(buf);
    return(NULL);
  }
  if (state->compress) {
    /* Decompress the query */
    char *data;
    z_stream *d_stream = NULL;	/* decompression stream */
    
    if ((data = malloc(state->uncompressLen+1)) == NULL) {
      free(buf);
      return(NULL);
    }
    if (!(d_stream = InitUncompress())) { 
      fprintf(stderr,"InitUncompress err [getquery:rover.c]\n");
      free(buf);
      return(NULL);
    }
    i = count = 0;
    rc = UncompressBuffer(d_stream, buf, &count, data, &i,
			  state->uncompressLen+1, &count);
    if (rc == 0) {
      fprintf(stderr, "Uncompress error: incomplete query/request\n");
      free(buf);
      free(data);
      return(NULL);
    } else if (rc == 2) {
      /* Leftover data! */
      fprintf(stderr, "Uncompress error: leftover data\n");
      free(buf);
      free(data);
      return(NULL);
    } else if (rc == -1) {
      fprintf(stderr, "Uncompress error\n");
      free(buf);
      free(data);
      return(NULL);
    } else if (rc != 1) {
      fprintf(stderr, "Uncompress error: unknown rc %d\n", rc);
      free(buf);
      free(data);
      return(NULL);
    } 
    /* rc == 1 normal end */
    free(buf);
    if (count != state->uncompressLen) {
      fprintf(stderr, "getquery got short decompress (%d; not %d)\n", 
	      count, state->uncompressLen);
      free(data);
      return(NULL);
    }
    buf = data;
  }

  buf[count] = '\0';
/*  fprintf(stderr, "Buffer is (%d) `%s'\n", (int) strlen(buf), buf);  */
  return(buf);
}


PUBLIC void main(int argc, char **argv)
{
  int i;
  char *buf, *name;

#ifdef EXEC
  exit(0);	/* Only measure the time to exec */
#endif /* EXEC */

  ServerName = argv[0];
  state = &(States[0]);
#ifdef TIMER
  gettimeofday(&ServerStartTv, &Tz);
  fprintf(stderr,"\nrover.c:main:%s",ctime((const time_t *)&(StartTv.tv_sec)));
#endif /* TIMER */

  Tcl_InitHashTable(&Info, TCL_STRING_KEYS);	/* User Info htable */
  /* Prevent apps from causing problems */
  i = dup(fileno(stdin));
  state->Stdin = fdopen(i, "r");
  i = dup(fileno(stdout));
  state->Stdout = fdopen(i, "w");
  fclose(stdin);
  fclose(stdout);
  
  if ((buf = getenv("CONTENT_LENGTH")))
    state->ContentLength = atoi(buf);
  else
    state->ContentLength = 0;
  state->ContentType = getenv("CONTENT_TYPE");
  if ((buf = getenv("COMPRESS")))
    state->compress = atoi(buf);
  else
    state->compress = 0;
  if ((buf = getenv("LENGTH")))
    state->uncompressLen = atoi(buf);
  else
    state->uncompressLen = 0;

  if (setjmp(state->RoverEnv) == 0) {
    /* Normal, direct call */
  } else {
    /* Opps, an errorExit occurred */
    return;
  }

  RoverRootDirsInit();

  initState(state);
  state->inuse = 1;
  state->first = 1;
 
  if (!(name = getenv("REQUEST_METHOD"))) {
    state->status = 400;
    errorExit("No REQUEST_METHOD specified");
  }
  if (strcmp(name,"POST")) {
    state->status = 500;
    errorExit("REQUEST_METHOD isn't expected POST type");
  }
  if (!strcmp(name,"POST")) {
    if (!state->ContentType) {
      state->status = 400;
      errorExit("Server called without a CONTENT_TYPE");
    }
    if (strcmp(state->ContentType, "application/x-rover")) {
      sprintf(line, "Got CONTENT_TYPE: %s; not application/x-rover",
	      state->ContentType);
      state->status = 400;
      errorExit(line);
    }
    state->Query = getquery(state->Stdin);
  }

  process_query(state);
  if (!state->batch)
    rovermain(state, argc, argv);
  else {
    for (i = 0; i < MAXSTATES; i++)
      if (States[i].inuse) {
	state = &(States[i]);	
	rovermain(state, argc, argv);
      }
  }
}

PUBLIC void saveState(ServerState *s)
{
}

PUBLIC void LogChange(char *action)
{
  fprintf(stderr, "WARNING! Logging not enabled for CGI server\n");
}

/*
 * Construct and send a status header to the client, followed
 *   by the output buffer and any dependency information.
 */
PUBLIC int sendServerResponse(int *extra, int noCompress)
{
  int n, rc = 0;
  z_stream *c_stream = NULL;	/* Compression stream */

  if (standalone && (state->loggedOutput == 1)&& (ServerRecovering != READLOG))
    stableLogOutput();

  generateServerHeader(state);

  if (state->openConnection) {
    /* Open a connection to the client here */
    if (!state->httpdHost || !state->port) {
      fprintf(stderr, "Missing httpdHost (%s) or port (%d) information\n",
	      state->httpdHost, state->port);
      return(0);
    }
    fprintf(stderr, "Opening HTTP connection to the client\n");
    if ((state->tcp->sock = openTCP(state->httpdHost, state->port)) == -1) {
      fprintf(stderr,"openTCP to %s:%d failed\n",state->httpdHost,state->port);
      exit(-1);
    }
    state->Stdin = fdopen(state->tcp->sock, "r");/* Convert to streams. */
    state->Stdout = fdopen(state->tcp->sock, "w");
    if (!state->Stdin || !state->Stdout) {
      fprintf(stderr, "fdopen(state->tcp->sock) failed");
      exit(-1);
    }
  }

  /* Output the header */
  fputs(state->Header, state->Stdout);
  fflush(state->Stdout);

  if (!state->OutSize && !state->DVSize) return(rc);

  if (!noCompress && state->compress) {
    if (!(c_stream = InitCompress(state->compress))) {
      fprintf(stderr, "InitCompress error");
      exit(-1);
    }

    n = 0;
    if (Compress(c_stream, state->Stdout, state->output->buffer,state->OutSize,
		 output,2048,&n) != Z_OK) {
      fprintf(stderr, "Compress error");
      exit(-1);
    }
    rc = n;
    if (Compress(c_stream, state->Stdout, state->DVbuf, state->DVSize, output,
		 2048, &n) != Z_OK) {
      fprintf(stderr, "Compress error");
      exit(-1);
    }
    rc += n;
    *extra = 0;

    if (EndCompress(c_stream, state->Stdout, output, 2048, &n) != Z_OK) {
      fprintf(stderr, "EndCompress error");
      exit(-1);
    }
    rc += n;
  } else {
    rc = fwrite(state->output->buffer, sizeof(char), state->OutSize,
		state->Stdout);
    *extra = fwrite(state->DVbuf, sizeof(char), state->DVSize, state->Stdout);
  }

  fflush(state->Stdout);
  return(rc);
}

