 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rovertcl.c
 *
 * Description: Rover Server Tcl interface module.
 *
 *
 * Global Functions: None
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1996-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include "rover.h"
#include "rovertcp.h"
#include "database.h"
#include <sys/param.h>
#include "RoverEscape.h"

PUBLIC Tcl_Interp *globalTclInterp = NULL;
PUBLIC int ServerUseStableVars = 1;
PUBLIC int NullTrace = 0;
PUBLIC int ServerDoFail = 0;

PRIVATE char line[2050];

#ifdef STANDALONE
PRIVATE char command[256];
#endif /* STANDALONE */

/*
 * Delete the current interpreter (for the standalone server, also
 *   delete the saved copy).
 */
PUBLIC void deleteTclInterp()
{
#ifdef STANDALONE
  UserData *data;

  sprintf(command, "UserData_%s@%s_%s", state->ClientUser, state->rhost,
	  state->Service);
  hentry = Tcl_FindHashEntry(&Info, command);
  if (hentry && (data =(UserData *)Tcl_GetHashValue(hentry)) && data->interp) {
    if (globalTclInterp != data->interp) {
      fprintf(stderr,
	      "WARNING: data->interp != globalTclInterp [deleteTclInterp:rover.c]\n");
    }
    data->interp = NULL;
    Tcl_DeleteHashEntry(hentry);
  }
#endif /* STANDALONE */
  if (globalTclInterp) {
    Tcl_DeleteInterp(globalTclInterp);
    globalTclInterp = NULL;
  }
}

/* Rover_Abort */
PUBLIC int Rover_Abort(ClientData clientData, Tcl_Interp *interp, int argc,
		       char *argv[])
{
  fprintf(stderr, "Abort: ServerDoFail %d ServerRecovering %d\n", ServerDoFail, ServerRecovering);
  if (ServerDoFail && (ServerRecovering == NORMALOP)) abort();
  return TCL_OK;
}
/* Rover_outstrInit [<size>] */
PUBLIC int Rover_outstrInit(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[])
{
  int size;

  if (argc == 2) {
    if (Tcl_GetInt(interp, argv[1], &size) != TCL_OK) {
      sprintf(interp->result, "integer expected: %s [<size>]", argv[0]);
      return TCL_ERROR;
    }
    outstrInit(size);
  } else
    outstrInit(INITSIZE);
  return TCL_OK;
}

/* Rover_emitData */
PUBLIC int Rover_emitData(ClientData clientData, Tcl_Interp *interp, int argc,
			  char *argv[])
{
  emitData();
  return TCL_OK;
}


/* Rover_emitLog */
PUBLIC int Rover_emitLog(ClientData clientData, Tcl_Interp *interp, int argc,
			  char *argv[])
{
  emitLog();
  return TCL_OK;
}

/* Rover_outstrD [<arg1> ... <argn>] */
PUBLIC int Rover_outstrD(ClientData clientData, Tcl_Interp *interp, int argc,
			 char *argv[])
{
  int i;
  for (i = 1; i < argc; i++)
    outstrD(argv[i]);
  return TCL_OK;
}

/* Rover_escaped_outstrD [<arg1> ... <argn>] */
PUBLIC int Rover_escaped_outstrD(ClientData clientData, Tcl_Interp *interp,
				 int argc, char *argv[])
{
  int i;
  char *rv;

  for (i = 1; i < argc; i++) {
    rv = RoverEscapeTcl(argv[i]);
    outstrD(rv);
    if (rv != argv[i]) free(rv);
  }
  return TCL_OK;
}

/* Rover_outstrL <timestamp> <host> <logID> <status> <operation> */
PUBLIC int Rover_outstrL(ClientData clientData, Tcl_Interp *interp, int argc,
			 char *argv[])
{
  RoverOpStatus status;
  int logID, i, timestamp;
  Tcl_DString oper;

  if (argc < 5) {
    interp->result =
      "wrong # args: Rover_outstrL <timestamp> <host> <logID> <status> <operation>";
    return TCL_ERROR;
  }

  if ((Tcl_GetInt(interp, argv[1], &timestamp) != TCL_OK) ||
      (Tcl_GetInt(interp, argv[3], &logID) != TCL_OK) ||
      (Tcl_GetInt(interp, argv[4],  (int *)&status) != TCL_OK)) {
    sprintf(interp->result,
	    "integer expected: %s <timestamp> <host> <logID> <status> <operation>",
	    argv[0]);
      return TCL_ERROR;
  }

  Tcl_DStringInit(&oper);
  for (i = 5; i < argc; i++) {
      Tcl_DStringAppend(&oper, argv[i], -1);
      Tcl_DStringAppend(&oper, " ", -1);
  }
  outstrL((time_t) timestamp, argv[2], logID, status, Tcl_DStringValue(&oper));
  Tcl_DStringFree(&oper);
  return TCL_OK;
}

/* Rover_escaped_outstrL <timestamp> <host> <logID> <status> <operation> */
PUBLIC int Rover_escaped_outstrL(ClientData clientData, Tcl_Interp *interp,
				 int argc, char *argv[])
{
  RoverOpStatus status;
  int logID, i, timestamp;
  Tcl_DString oper;
  char *rv;

  if (argc < 5) {
    interp->result =
      "wrong # args: Rover_outstrL <timestamp> <host> <logID> <status> <operation>";
    return TCL_ERROR;
  }

  if ((Tcl_GetInt(interp, argv[1], &timestamp) != TCL_OK) ||
      (Tcl_GetInt(interp, argv[3], &logID) != TCL_OK) ||
      (Tcl_GetInt(interp, argv[4],  (int *)&status) != TCL_OK)) {
    sprintf(interp->result,
	    "integer expected: %s <timestamp> <host> <logID> <status> <operation>",
	    argv[0]);
      return TCL_ERROR;
  }

  Tcl_DStringInit(&oper);
  for (i = 5; i < argc; i++) {
      Tcl_DStringAppend(&oper, argv[i], -1);
      Tcl_DStringAppend(&oper, " ", -1);
  }
  rv = RoverEscapeTcl(Tcl_DStringValue(&oper));
  outstrL((time_t) timestamp, argv[2], logID, status, rv);
  Tcl_DStringFree(&oper);
  if (rv != Tcl_DStringValue(&oper)) free(rv);
  return TCL_OK;
}

/* Rover_outstrC [<arg1> ... <argn>] */
PUBLIC int Rover_outstrC(ClientData clientData, Tcl_Interp *interp, int argc,
			 char *argv[])
{
  int i;
  for (i = 1; i < argc; i++)
    outstrC(argv[i]);
  return TCL_OK;
}

/* Rover_escaped_outstrC [<arg1> ... <argn>] */
PUBLIC int Rover_escaped_outstrC(ClientData clientData, Tcl_Interp *interp,
				 int argc, char *argv[])
{
  int i;
  char *rv;

  for (i = 1; i < argc; i++) {
    rv = RoverEscapeTcl(argv[i]);
    outstrC(rv);
    if (rv != argv[i]) free(rv);
  }
  return TCL_OK;
}

/* Rover_no_cache */
PUBLIC int Rover_no_cache(ClientData clientData, Tcl_Interp *interp, int argc,
			  char *argv[])
{
  ServerState *s = (ServerState *) clientData;

  s->Cacheable = ROVER_NO_CACHE;
  return TCL_OK;
}

/*  Rover_Escape <string> */
PUBLIC int Rover_Escape(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[])
{
  char *rv;

  if (argc != 2) {
    interp->result = "wrong # args: Rover_Escape <string>";
    return TCL_ERROR;
  }
  if ((rv = RoverEscape(argv[1], URL_XALPHAS)) == NULL) {
    interp->result =  "Rover_Escape failed";
    return TCL_ERROR;
  }
  if (rv == argv[1]) rv = strdup(rv);	/* We must return a malloc'd string */
  interp->result = rv;
  interp->freeProc = (void (*)(char*))free;
  return TCL_OK;
}

/*  Rover_TclEscape <string> */
PUBLIC int Rover_TclEscape(ClientData clientData, Tcl_Interp *interp, int argc,
			   char *argv[])
{
  char *rv;

  if (argc != 2) {
    interp->result ="wrong # args: Rover_TclEscape <string>";
    return TCL_ERROR;
  }

  rv = RoverEscapeTcl(argv[1]);
  if (rv == argv[1]) rv = strdup(rv);	/* We must return a malloc'd string */
  interp->freeProc = (void (*)(char*))free;
  interp->result = rv;
  return TCL_OK;
}
/* Rover_Unescape <string> */
PUBLIC int Rover_Unescape(ClientData clientData, Tcl_Interp *interp, int argc,
			  char *argv[])
{
  char *rv;

  if (argc != 2) {
    interp->result = "wrong # args: Rover_Unescape <string>";
    return TCL_ERROR;
  }

  rv = RoverUnescape(argv[1]);
  if (rv == argv[1]) rv = strdup(rv);	/* We must return a malloc'd string */
  interp->result = rv;
  interp->freeProc = (void (*)(char*))free;
  return TCL_OK;
}

/* Rover_DbOpenWrite <filename> [<mode>] */
PUBLIC int Rover_DbOpenWrite(ClientData clientData, Tcl_Interp
*interp, int argc, char *argv[]) 
{
  int mode = 0644;
  ServerState *s = (ServerState *) clientData;

  if ((argc != 3) && (argc != 2)) {
    sprintf(interp->result, "wrong # args: %s <filename> [<mode>]", argv[0]);
    return TCL_ERROR;
  }
  if (argc == 3) 
    if (Tcl_GetInt(interp, argv[2], &mode) != TCL_OK) {
      sprintf(interp->result, "integer expected: %s <filename> [<mode>]",
	      argv[0]);
      return TCL_ERROR;
    }
  db_openWrite(argv[1], &(s->currdb), NULL, mode);
  return TCL_OK;
}

/* Rover_DbOpenRead <filename> <mode> */
PUBLIC int Rover_DbOpenRead(ClientData clientData, Tcl_Interp *interp,
int argc, char *argv[]) 
{
  int mode;
  ServerState *s = (ServerState *) clientData;

  if ((argc != 3) && (argc != 2)) {
    sprintf(interp->result, "wrong # args: %s <filename> [<mode>]", argv[0]);
    return TCL_ERROR;
  }
  if (argc == 3) 
    if (Tcl_GetInt(interp, argv[2], &mode) != TCL_OK) {
      sprintf(interp->result, "integer expected: %s <filename> [<mode>]",
	      argv[0]);
      return TCL_ERROR;
    }
  db_openRead(argv[1], &(s->currdb), NULL, mode);
  return TCL_OK;
}

/* Rover_DbReopenWrite */
PUBLIC int Rover_DbReopenWrite(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[]) 
{
  ServerState *s = (ServerState *) clientData;

  if (argc != 1) {
    sprintf(interp->result, "wrong # args: %s", argv[0]);
    return TCL_ERROR;
  }
  db_reopenWrite(&(s->currdb));
  return TCL_OK;
}

/* Rover_DbReopenRead */
PUBLIC int Rover_DbReopenRead(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[]) 
{
  ServerState *s = (ServerState *) clientData;

  if (argc != 1) {
    sprintf(interp->result, "wrong # args: %s", argv[0]);
    return TCL_ERROR;
  }
  db_reopenRead(&(s->currdb));
  return TCL_OK;
}

/* Rover_DbClose */
PUBLIC int Rover_DbClose(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[]) 
{
  ServerState *s = (ServerState *) clientData;

  if (argc != 1) {
    sprintf(interp->result, "wrong # args: %s", argv[0]);
    return TCL_ERROR;
  }
  db_close(&(s->currdb));
  return TCL_OK;
}

/* Rover_DbDelete <obj> */
PUBLIC int Rover_DbDelete(ClientData clientData, Tcl_Interp *interp, int argc,
			  char *argv[])
{
  datum rq;
  int rc;
  ServerState *s = (ServerState *) clientData;

  /* Delete an object from the database */
  if (argc != 2) {
    interp->result = 
      "wrong # args: Rover_DbDelete <obj>";
    return TCL_ERROR;
  }

  rq.dptr = argv[1];
  rq.dsize = strlen(argv[1])+1;
  rc = gdbm_delete(s->currdb.db, rq);	/* Delete the object*/

  if ((interp->result = malloc(10)) == NULL) {
    interp->result = "Memory allocation failed";
    return TCL_ERROR;
  }
  interp->freeProc = (void (*)(char*))free;
  sprintf(interp->result, "%d", rc);
  return TCL_OK;
}

/* Rover_GetAccessTime <obj> */
PUBLIC int Rover_GetAccessTime(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  ServerState *s = (ServerState *) clientData;

  if (argc != 2) {
    sprintf(interp->result, "wrong # args: %s <obj>", argv[0]);
    return TCL_ERROR;
  }
  sprintf(interp->result, "%ld", db_getAccessTime(&(s->currdb), argv[1]));
  
  return TCL_OK;
}

/* Rover_SetAccessTime <obj> <stamp> */
PUBLIC int Rover_SetAccessTime(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  time_t stamp;
  ServerState *s = (ServerState *) clientData;

  if (argc != 3) {
      sprintf(interp->result, "wrong # args: %s <obj> <stamp>", argv[0]);
    return TCL_ERROR;
  }

  if (Tcl_GetInt(interp, argv[2], (int *)&stamp) != TCL_OK) {
      sprintf(interp->result, "integer expected: %s <obj> <stamp>", argv[0]);
      return TCL_ERROR;
  }

  db_setAccessTime(&(s->currdb), argv[1], stamp);
  
  return TCL_OK;
}

/* Rover_AddLog <obj> <time> <status> <operation> */
PUBLIC int Rover_AddLog(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[])
{
  int ltime;
  ServerState *s = (ServerState *) clientData;

  if (argc != 5) {
    interp->result = 
      "wrong # args: Rover_AddLog <obj> <time> <status> <operation>";
    return TCL_ERROR;
  }
  if (s->currdb.db == NULL) {
    interp->result = "No database open";
    return TCL_ERROR;
  }
  
  if (Tcl_GetInt(interp, argv[2], &ltime) != TCL_OK) {
    sprintf(interp->result,
	    "integer expected: %s <obj> <time> <status> <operation>", argv[0]);
    return TCL_ERROR;
  }
  log_add(&(s->currdb), argv[1], (time_t)ltime, s->rhost, s->LogID,
	  (!strcasecmp(argv[3], "commit") ? OP_COMMIT: OP_ABORT), argv[4],
	  strlen(argv[4])+1);

  return TCL_OK;
}

/* Rover_outstrLSuffix <obj> <stamp> */
PUBLIC int Rover_outstrLSuffix(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  time_t stamp;
  ServerState *s = (ServerState *) clientData;

  if (argc != 3) {
      sprintf(interp->result,"wrong # args: %s <obj> <stamp>", argv[0]);
    return TCL_ERROR;
  }

  if (Tcl_GetInt(interp, argv[2], (int *)&stamp) != TCL_OK) {
      sprintf(interp->result, "integer expected: %s <obj> <stamp>", argv[0]);
      return TCL_ERROR;
  }

  log_suffix(&(s->currdb), argv[1], stamp, log_outstrSuffix);
  
  return TCL_OK;
}

/* Rover_CTime <time> */
PUBLIC int Rover_CTime(ClientData clientData, Tcl_Interp *interp, int argc,
		       char *argv[])
{
  int i = 0;
  time_t t;

  if (argc != 2) {
    interp->result = "wrong # args: Rover_CTime <time>";
    return TCL_ERROR;
  }
  if (Tcl_GetInt(interp, argv[1], &i) != TCL_OK) {
    sprintf(interp->result, "integer expected: %s <time>", argv[0]);
    return TCL_ERROR;
  }
  t = (time_t) i;
  interp->result = ctime(&t);
  interp->result[strlen(interp->result)-1] = '\0';
  return TCL_OK;
}

/* Rover_setRecoveryProc <procedure> */
PUBLIC int Rover_setRecoveryProc(ClientData clientData, Tcl_Interp *interp,
				 int argc, char *argv[])
{
#ifndef STANDALONE
  interp->result = "Forked Rover server doesn't support recovery procedures.";
  return TCL_ERROR;
#else
  ServerState *s = (ServerState *) clientData;

  if (argc != 2) {
    interp->result = "wrong # args: Rover_setRecoveryProc <procedure>";
    return TCL_ERROR;
  }
  s->recoverProc = copyString(argv[1]);
  if (ServerNoLog) return TCL_OK;
  fprintf(ServerLogFile, "ROVER_LOG_RECOVER %d %s\n", s->ServerSeq, argv[1]);
  fflush(ServerLogFile);
  if (ServerHardFlush) fsync(ServerLogFD);
  return TCL_OK;
#endif /* !STANDALONE */
}

#ifdef STANDALONE
PUBLIC char *stableTraceCallback(ClientData clientData, Tcl_Interp *interp,
				 char *name1, char *name2, int flags)
{
  char *value;
  int size, rc;
  static int count = 0;
  ServerState *s = (ServerState *) clientData;

  if (NullTrace) return NULL;
#if 0
  {
    int i = atoi(Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY));
    if ((ServerRecovering == NORMALOP) && (i == 20)) abort();
  }
#endif

  if (name2 == NULL)
    strcpy(line, name1);
  else
    sprintf(line, "%s(%s)", name1, name2);
  if (flags & TCL_TRACE_DESTROYED) {
    fprintf(stderr, "Rover stable variable %s unset, untracing\n", line);
    Tcl_UntraceVar2(interp, name1, name2, TCL_STABLE_FLAGS,
		    stableTraceCallback, (ClientData) NULL);
  } else if (flags & TCL_INTERP_DESTROYED) {
    fprintf(stderr, "Rover Tcl interpreter containing %s destroyed\n", line);
  } else {
    value = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
/*    fprintf(stderr, "Rover stable var %s changed: %s\n", line, value); */

fprintf(stderr, "Trace count is %d\n", ++count);
    size = strlen(value)+1;
#if 0
    if (ServerDoFail) {
      if ((ServerRecovering == NORMALOP) && (size == 475)) abort();
    }
#endif
    if (ServerNoLog) return NULL;
    fprintf(ServerLogFile, "ROVER_LOG_STABLE_VAR %d %s %d\n", s->ServerSeq,
	    line, size);
    if (size)
      if ((rc = fwrite(value, sizeof(char), size, ServerLogFile)) != size) {
	s->status = 500;
	sprintf(line, "Bad write size %d [Rover_stable:rovertcl.c]\n", rc);
	errorExit(line);
      }
    fputc('\n', ServerLogFile);
    fflush(ServerLogFile);
    if (ServerHardFlush) fsync(ServerLogFD);
  }
  return NULL;
}

#endif /* STANDALONE */

/*
 * Restore the stable variables of a request, setting up traces.
 */
PUBLIC void restoreStableVars()
{
#ifdef STANDALONE

  int i;

  for (i = 0; i < MAXVAR; i++) {
    if (state->vars[i].inuse) {
      if (state->vars[i].value) {
	/* Set the variable before we trace it */
	Tcl_SetVar(globalTclInterp, state->vars[i].name, state->vars[i].value,
		   TCL_GLOBAL_ONLY);
      }
      Tcl_TraceVar(globalTclInterp, state->vars[i].name, TCL_STABLE_FLAGS,
		   stableTraceCallback, (ClientData) NULL);
    }
  }
#endif /* STANDALONE */
}

/* Rover_stable <name> [<value>] */
PUBLIC int Rover_stable(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[])
{
#ifndef STANDALONE
  interp->result = "Forked Rover server doesn't support stable variables.";
  return TCL_ERROR;
#else
  ServerState *s = (ServerState *) clientData;
  int rc, size, i, found;

  if ((argc < 2) || (argc > 3)) {
    interp->result = "wrong # args: Rover_stable <name> [<value>]";
    return TCL_ERROR;
  }
  if (!ServerUseStableVars) {
    if (argc == 3) {
      /* Set the variable before we trace it */
      Tcl_SetVar(interp, argv[1], argv[2], TCL_GLOBAL_ONLY);
    }
    return TCL_OK;
  }
  /* Search for the stable var (to prevent duplicates) and an empty slot */
  for (rc = -1, found = i = 0; i < MAXVAR; i++)
    if (s->vars[i].inuse) {
      if (!strcmp(s->vars[i].name, argv[1])) {
	found = 1;
	break;
      }
    } else if (rc == -1) rc = i;
  if (!found) {
    /* Not found, good */
    if (rc != -1) {
      /* Got an empty slot, also good */
      i = rc;
      s->vars[i].inuse = 1;
      s->vars[i].name = copyString(argv[1]);
      if (argc == 3) s->vars[i].value = copyString(argv[2]);
      else s->vars[i].value = NULL;
    } else {
      /* No slots, bad */
      fprintf(stderr, "Out of stable var slots [Rover_stable:rovertcl.c]\n");
      interp->result = "Out of stable var slots";
      return TCL_ERROR;
    }
  } else {
    /* Found dcl, bad */
    interp->result = "WARNING! Duplicate server stable variable declaration";
    return TCL_OK;
  }

  if (argc == 3) {
    /* Set the variable before we trace it */
    Tcl_SetVar(interp, argv[1], argv[2], TCL_GLOBAL_ONLY);
  }
  Tcl_TraceVar(globalTclInterp, argv[1], TCL_STABLE_FLAGS, stableTraceCallback,
	       (ClientData) NULL);
  if (ServerNoLog) return TCL_OK;
  if (argc == 2) {
    fprintf(ServerLogFile, "ROVER_LOG_STABLE_DCL %d %s 0\n", s->ServerSeq,
	    argv[1]);
  } else {
    size = strlen(argv[2])+1;
    fprintf(ServerLogFile, "ROVER_LOG_STABLE_DCL %d %s %d\n", s->ServerSeq,
	    argv[1], size);
    if (size)
      if ((rc = fwrite(argv[2], sizeof(char), size, ServerLogFile)) != size) {
	s->status = 500;
	sprintf(line, "Bad write size %d [Rover_stable:rovertcl.c]\n", rc);
	errorExit(line);
      }
  }
  fputc('\n', ServerLogFile);
  fflush(ServerLogFile);
  if (ServerHardFlush) fsync(ServerLogFD);
  return TCL_OK;
#endif /* !STANDALONE */
}

#define MAXTIMERS 1000
PRIVATE struct timeval time_start[MAXTIMERS], time_stop[MAXTIMERS];  /* for real time */
PRIVATE int timecount = 0;

/* Rover_StartTimer */
PRIVATE int Rover_StartTimer(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[])
{
  char buf[132];

  if (argc != 1) {
    interp->result ="wrong # args: Rover_StartTimer";
    return TCL_ERROR;
  }

  if (gettimeofday(&time_start[timecount], (struct timezone *) 0) < 0) {
    interp->result = "t_start: gettimeofday() error";
    return TCL_ERROR;
  }
  sprintf(buf, "%d", timecount);
  timecount = (timecount+1)%MAXTIMERS;
  Tcl_SetResult(interp, buf, TCL_VOLATILE);
  return TCL_OK;
}

/* Rover_EndTimer <ctr> */
PRIVATE int Rover_EndTimer(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[])
{
  char buf[132];
  int count;
  double start, stop, seconds;

  if (argc != 2) {
    interp->result ="wrong # args: Rover_EndTimer <ctr> ";
    return TCL_ERROR;
  }

  if (Tcl_GetInt(interp, argv[1], &count) != TCL_OK) {
    sprintf(interp->result, "integer expected: %s <ctr>", argv[0]);
    return TCL_ERROR;
  }
  if (count >=  MAXTIMERS) {
    interp->result ="Not a valid timer identifier";
    return TCL_ERROR;
  }

  if (gettimeofday(&time_stop[count], (struct timezone *) 0) < 0) {
    interp->result = "t_stop: gettimeofday() error";
    return TCL_ERROR;
  }

  start = ((double) time_start[count].tv_sec) * 1000000.0
    + time_start[count].tv_usec;
  stop = ((double) time_stop[count].tv_sec) * 1000000.0
    + time_stop[count].tv_usec;
  seconds = (stop - start) / 1000000.0;

  sprintf(buf, "%.4e", seconds);
  Tcl_SetResult(interp, buf, TCL_VOLATILE);
  return TCL_OK;
}

/*
 * Load a tcl file into the interpreter.
 */
PUBLIC void Tcl_LoadLibrary(char *name, ServerState *s)
{
  struct stat sbuf;
  char dbuf[32], *value;

  if (stat(name, &sbuf)) {
    sprintf(line,
	    "Load Lib %s fail: file not found [Tcl_LoadLibrary:rovertcl.c]\n",
	    name);
    s->status = 500;
    errorExit(line);
  }
  sprintf(line, "TLL__%s", name);
  value = Tcl_GetVar(globalTclInterp, line, TCL_GLOBAL_ONLY);
  if (!value || (atol(value) < sbuf.st_mtime)) {
    /* Either the library wasn't loaded or an old version was loaded */
    sprintf(dbuf, "%lu", (unsigned long) sbuf.st_mtime);
    Tcl_SetVar(globalTclInterp, line, dbuf, TCL_GLOBAL_ONLY);
    if (Tcl_EvalFile(globalTclInterp, name) != TCL_OK) {
      sprintf(line ,
	      "Load Library %s failed: TCL EVAL ERR: %s [Tcl_LoadLibrary]\n",
	      name, globalTclInterp->result);
      s->status = 500;
      errorExit(line);
    }
  }
}

/* 
 * Create a Tcl interpreter, or if running standalone, look for an
 *   interpreter in the tuple consisting of: <Host, user, service>.
 *   If an interpreter is created, run the local config file
 *   (which among other things should remove any commands we don't want
 *   downloaded programs to use), and then install some Rover commands.
 *
 * Returns: 0, if a new interpreter was created; 1, if an existing
 *	    interpreter was found; -1, if an error occurred.
 */
#define TCL_FLAGS 		(TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)
PUBLIC int createTclInterp(ServerState *s)
{
  char initFile[MAXPATHLEN+1], buf[32];
#ifdef STANDALONE
  int new;
  UserData *data = NULL;

  sprintf(command, "UserData_%s@%s_%s", s->ClientUser, s->rhost, s->Service);
  hentry = Tcl_FindHashEntry(&Info, command);
  if (hentry && (data =(UserData *)Tcl_GetHashValue(hentry)) && data->interp) {
    globalTclInterp =  data->interp;
    if (RoverDebug) fprintf(stderr, "Fast path [createTclInterp:rovertcl.c]\n");

    /* Reset a few environment variables */
    Tcl_SetVar2(globalTclInterp, "env", "USER", s->ClientUser, TCL_FLAGS);
    if (s->Home)
      Tcl_SetVar2(globalTclInterp, "env", "HOME", s->Home, TCL_FLAGS);
    
    Tcl_SetVar(globalTclInterp, "rhost", s->rhost, TCL_FLAGS);
    sprintf(buf, "%d", s->LogID);
    Tcl_SetVar(globalTclInterp, "LogID", buf, TCL_FLAGS);
    Tcl_SetVar(globalTclInterp, "ClientUser", s->ClientUser, TCL_FLAGS);
    sprintf(line, "%lu", s->UserTime);
    Tcl_SetVar(globalTclInterp, "UserTime", line, TCL_FLAGS);

    return(1);		/* Found an existing interpreter */
  }
  if (globalTclInterp) {
      fprintf(stderr,
	      "WARNING: undifferentiated globalTclInterp exists [createTclInterp:rover.c]\n");
      return(1);	/* Found an existing interpreter */
  }

#endif /* STANDALONE */

  /* Create an interpreter */
  globalTclInterp = Tcl_CreateInterp();
  if (Tcl_Init(globalTclInterp) != TCL_OK) {
    fprintf(stderr, "Error from Tcl_Init: %s [createTclInterp:rovertcl.c]\n",
	     globalTclInterp->result);
    return(-1);
  }

#define TCL_MACH_VAR 		"Rover__Env__machineType"
#define TCL_NET_VAR 		"Rover__Env__netStatus"
#define TCL_SERVER_ID_VAR 	"Rover__Env__serverID"

  /* Set a variable for server ID */
  sprintf(line, "%u", ServerID);
  if (Tcl_SetVar(globalTclInterp, TCL_SERVER_ID_VAR, line, TCL_FLAGS)== NULL) {
    fprintf(stderr, "SetVar error: %s [createTclInterp:rovertcl.c]\n",
	     globalTclInterp->result);
    return(-1);
  }

  /* Set a variable for current network status */
  Tcl_SetVar(globalTclInterp, TCL_NET_VAR, "2", TCL_FLAGS);

  /* Setup the interpreter's initial environment */
#ifdef ultrix
  Tcl_SetVar(globalTclInterp, TCL_MACH_VAR, "dec", TCL_FLAGS);
#elif defined(linux)
  Tcl_SetVar(globalTclInterp, TCL_MACH_VAR, "linux", TCL_FLAGS);
#elif defined(sparc)
  Tcl_SetVar(globalTclInterp, TCL_MACH_VAR, "sun4", TCL_FLAGS);
#endif

  /* Set a few environment variables */
  Tcl_SetVar2(globalTclInterp, "env", "USER", s->ClientUser, TCL_FLAGS);
  if (s->Home)
    Tcl_SetVar2(globalTclInterp, "env", "HOME", s->Home, TCL_FLAGS);
  Tcl_SetVar(globalTclInterp, "rhost", s->rhost, TCL_FLAGS);
  sprintf(buf, "%d", s->LogID);
  Tcl_SetVar(globalTclInterp, "LogID", buf, TCL_FLAGS);
  Tcl_SetVar(globalTclInterp, "ClientUser", s->ClientUser, TCL_FLAGS);
  sprintf(line, "%lu", s->UserTime);
  Tcl_SetVar(globalTclInterp, "UserTime", line, TCL_FLAGS);
  if (Tcl_SetVar(globalTclInterp, "RoverRoot", RoverRoot[RoverRootID],
		 TCL_FLAGS) == NULL) {
    fprintf(stderr, "SetVar error: %s [createTclInterp:rovertcl.c]\n",
	    globalTclInterp->result);
    return(-1);
  }

  /* Now, add our new commands */
  RoverCreateTclCmd(Rover_Abort, s);
  RoverCreateTclCmd(Rover_AddLog, s);
  RoverCreateTclCmd(Rover_CTime, s);
  RoverCreateTclCmd(Rover_DbClose, s);
  RoverCreateTclCmd(Rover_DbDelete, s);
  RoverCreateTclCmd(Rover_DbOpenRead, s);
  RoverCreateTclCmd(Rover_DbOpenWrite, s);
  RoverCreateTclCmd(Rover_DbReopenRead, s);
  RoverCreateTclCmd(Rover_DbReopenWrite, s);
  RoverCreateTclCmd(Rover_EndTimer, s);
  RoverCreateTclCmd(Rover_Escape, s);
  RoverCreateTclCmd(Rover_GetAccessTime, s);
  RoverCreateTclCmd(Rover_SetAccessTime, s);
  RoverCreateTclCmd(Rover_StartTimer, s);
  RoverCreateTclCmd(Rover_TclEscape, s);
  RoverCreateTclCmd(Rover_Unescape, s);
  RoverCreateTclCmd(Rover_emitData, s);
  RoverCreateTclCmd(Rover_emitLog, s);
  RoverCreateTclCmd(Rover_escaped_outstrC, s);
  RoverCreateTclCmd(Rover_escaped_outstrD, s);
  RoverCreateTclCmd(Rover_escaped_outstrL, s);
  RoverCreateTclCmd(Rover_no_cache, s);
  RoverCreateTclCmd(Rover_outstrC, s);
  RoverCreateTclCmd(Rover_outstrD, s);
  RoverCreateTclCmd(Rover_outstrInit, s);
  RoverCreateTclCmd(Rover_outstrL, s);
  RoverCreateTclCmd(Rover_outstrLSuffix, s);
  RoverCreateTclCmd(Rover_setRecoveryProc, s);
  RoverCreateTclCmd(Rover_stable, s);


  /* Evaluate the initialization file. We should really check this file to
     make sure it is OK, since we're using it to configure the safe
     environment. The alternative would be to hard code it into Rover.
   */
  sprintf(initFile, "%s/init.tcl.obj", RoverRoot[RoverRootID]);
  if (Tcl_EvalFile(globalTclInterp, initFile) != TCL_OK) {
    fprintf(stderr, "TCL EVAL ERR: %s [createTclInterp:rovertcl.c]\n",
	     globalTclInterp->result);
    return(-1);
  }

#ifdef STANDALONE
  sprintf(command, "UserData_%s@%s_%s", s->ClientUser, s->rhost,
	  s->Service);
  if (data) {
    data->interp = globalTclInterp;
  } else {
    hentry = Tcl_CreateHashEntry(&Info, command, &new);
    if (!new) {				/* Check for duplicate names */
      sprintf(line,
	      "Duplicate Info hash table entry: %s [createTclInterp]\n", 
	      command);
      errorExit(line);
    }
    if ((data = (UserData *)malloc(sizeof(UserData))) == NULL) {
      fatalError("Server unable to allocate UserData buffer");
    }
    data->interp = globalTclInterp;
    Tcl_SetHashValue(hentry, (char *)data);	/* Save it */
#ifdef TCLDEBUG
    fprintf(stderr, "rover.c:createTclInterp: globalTclInterp %lX\n",
	    (long)globalTclInterp);
    fprintf(stderr, "rover.c:createTclInterp: hentry %lX\n", (long)hentry);
    fprintf(stderr, "rover.c:createTclInterp: data %lX\n", (long)data);
    fprintf(stderr, "rover.c:createTclInterp: data->interp %lX\n",
	    (long)data->interp);
#endif /* TCLDEBUG */
  }
#endif /* STANDALONE */

  return(0);
}


