 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rovertcp.h
 *
 * Description: Rover server-side TCP routines.
 *
 *
 * Global Functions: None
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef ROVER_TCP_H
#define ROVER_TCP_H

#include "tcp.h"
#include "HTUtils.h"
#include "HTAccess.h"
#include "HTError.h"
#include "HTThread.h"
#include "HTList.h"
#include "RoverEvent.h"
#include "RoverTCP.h"
#include "RoverZlib.h"

typedef enum _TCPPState {
  TCP_BEGIN = 0,
  TCP_NEED_REQUEST,
  TCP_GOT_REQUEST,
  TCP_READ_MIMEHDR,
  TCP_PROC_MIMEHDR,
  TCP_PREP_REQUEST,
  TCP_READ_BODY,
  TCP_SAVE_BODY,
  TCP_GOT_BODY,
  TCP_WAIT_MSGID,
  TCP_EXECUTE,
  TCP_CLEANUP,
  TCP_ABORT,
} TCPState;

typedef enum _TCPType {
  NOTYPE = 0,
  ROVER_REQ,
  HTTP_REQ,
} TCPType;

typedef enum _TCP_RESET {
  PARTIAL = 0,
  FULL = 1,
  ABORT = 2
} TCP_RESET;

struct _HashEntry {
  int size;
  char *data;
};
typedef struct _HashEntry HashEntry;

struct _StableVar {
  int inuse;
  char *name;
  char *value;
};
typedef struct _StableVar StableVar;

#define MAXVAR	64


struct _Buffers {
  int size;
  char *buffer;
  struct _Buffers *next;
};
typedef struct _Buffers Buffers;


struct _AuthInfo {
  int valid;		/* Flag indicating auth info has been validated */
  int Nobody;		/* Flag indicating whether user is type nobody */
  gid_t gid;		/* User's GID */
  uid_t uid;		/* User's UID */
  char *HomeEnv;	/* HOME= variable */
  char *UserEnv;	/* USER= variable */
};
typedef struct _AuthInfo AuthInfo;

#define HEADER_SIZE 256

struct _ServerState {
  int inuse;		/* Structure in-use bit */
  /* The following fields are saved in a stable log */
  int ServerSeq;	/* Server's seq. # (msg rcv order; not exec order) */
  unsigned int incarnation;	/* Client's incarnation */
  int msgID;		/* Client's message ID */
  char *ClientUser;	/* Client's user name */
  char *rhost;		/* Remote host name */
  char *httpdHost;	/* Remote host AM httpd name */
  int port;		/* Remote host AM httpd port */
  char *Service;	/* Service name */
  int ServiceID;	/* Client's Service ID */
  int LogID;		/* Client's Log ID */
  time_t UserTime;	/* Client's time for Object  */
  int UserServer;	/* Client's server ID for Object's time */
  char *Object;		/* Name of the object */
  char *command;	/* Client's request command */
  char *password;	/* Client's authenticator */
  char *code;		/* Client's code, if any */
  int size;		/* Size of the client code, if any */
  char *ContentType;	/* HTTP type of client request message */
  int ContentLength;	/* HTTP length of client request message */
  int retry;		/* Number of times client has sent request */

  /* The following fields are NOT saved in a stable log */
  int compress;		/* Compression level used for request */
  int uncompressLen;	/* Length of uncompressed data */
  TCPState state;
  TCPType reqType;
  TCPState nextState;
  HTEventCallBack *callback;
  NBTcp *tcp;		/* socket info for client */
  int count;		/* size of received request */
  int bufferOffset;
  z_stream *inflate;
  FILE *Stdin;		/* stdin for client */
  FILE *Stdout;		/* stdout for client */
  char *Query;		/* Complete request from client */
  int QuerySize;	/* Size of client request */
  void *db;		/* Database pointer */
  int pid;		/* Process id for request */
  char *recoverProc;	/* Tcl recovery procedure */
  StableVar vars[MAXVAR]; /* Stable variables */
  int order;		/* Execution order of the request */
  int openConnection;	/* Flag to indicate a connection to client's httpd */
  int keepalive;	/* Keep the connection open for the next request */
  int first;		/* First request in a batch of requests */
  int loggedOutput;	/* Log output buffer to stable storage */
  int CommandDone;	/* Flag indicating request is completed */
  int status;		/* HTTP request completion code */
  int batch;		/* Size of query batch */
  char *remoteIP;	/* Client IP address */
  char *remoteName;	/* Client hostname */

  AuthInfo auth;	/* User authentication information */
  jmp_buf RoverEnv;	/* Jump buffer for errors */
  char *Home;		/* User's home directory */
  char *FastHome;	/* User's home directory */
  Database currdb;	/* Current database */
  void *plugInData;	/* Private plugin information */

  time_t dvtime;	/* Dependency vector time, if one exists */
  int LogCount;		/* Number of records in a log suffix */
  char *LogStart;	/* Start of log suffix in output buffer */
  int EmittingLog;	/* Flag indicating server is emitting log records */
  int EmittingData;	/* Flag indicating server is emitting data */

  /* The following fields are saved when the output buffer is saved */
  char *outputBuffer;	/* Saved output buffer */
  RoverObjType ObjType; /* Type of return value */
  int OutSize;		/* Number of bytes in output buffer */
  int CodeSize;		/* Size of code section of output buffer */
  int DataSize;		/* Size of data section of output buffer */
  int Cacheable;	/* Flag to indicate cacheability of return value */
  int LogSize;		/* Size of log suffix */
  int DVSize;		/* Size of marshalled DV */
  char Header[HEADER_SIZE]; /* Header buffer */
  Buffers *output;	/* Output buffer */
  char DVbuf[32];	/* Buffer for marshalled DV */
};
typedef struct _ServerState ServerState;

struct _ClientMsgID {
  int nextID;		/* Next msg ID expected from client */
  unsigned int incarnation;	/* Client's current incarnation */
};
typedef struct _ClientMsgID ClientMsgID;

extern void outstrReleaseBuffer(Buffers *output);

extern int processHTTP(ServerState *s, int argc, char **argv, int SMTP);

extern int createTclInterp(ServerState *s);
extern void Tcl_LoadLibrary(char *name, ServerState *s);
extern void initState(ServerState *s);
extern int checkMsgID(char *user, char *host, int msgID, unsigned int incarnation, int retry);
extern void saveResult(int seq, int rc);
extern void saveState(ServerState *s);
extern void recover(char *name, ServerState *s);
extern void restoreStableVars();
extern int cleanLog();
extern void reset_state(ServerState *s);
extern void roverCloseLogs();
extern int roverInitLogFile(char *name, FILE **filep, int *fdp);

extern void rovermain(ServerState *s, int argc, char **argv);
extern void perform_command(ServerState *s);		/* In interface.c */
extern void recover_command(ServerState *s);		/* In interface.c */
extern void perform_nobody_command(ServerState *s);
extern void process_query(ServerState *s);
extern void reset_arp(ServerState *s, char *host);
extern void reset_auth();
extern void perform_auth_check(ServerState *s);
extern void roverResetAuth(ServerState *s);
extern void RoverSetAuth(ServerState *s);
extern void serverShowOptions();
extern int checkVerify(char *object, ServerState *s, RoverObjType utype);

extern ServerState *state;
extern ServerState States[];
extern ServerState *TCPStates[];
extern int ServerDoFail;

#define TCL_STABLE_FLAGS   (TCL_TRACE_WRITES|TCL_GLOBAL_ONLY|TCL_TRACE_UNSETS)

/* Macros to simplify setting up Tcl interpreter */
#define TCL_MACH_VAR 		"Rover__Env__machineType"
#define TCL_NET_VAR 		"Rover__Env__netStatus"
#define TCL_SERVER_ID_VAR 	"Rover__Env__serverID"
#define TCL_FLAGS 		(TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG)
#define RoverCreateTclCmd(CMD, STATE)	\
  Tcl_CreateCommand(globalTclInterp, #CMD, CMD, (ClientData) STATE, NULL);
#define RoverCreateTclCmdAlias(ALIAS, CMD, STATE)	\
  Tcl_CreateCommand(globalTclInterp, #ALIAS, CMD, (ClientData) STATE, NULL);


struct _UserData {
  Tcl_Interp *interp;	/* The tuple's Tcl interpreter */
};
typedef struct _UserData UserData;

#endif /* ROVER_TCP_H */
