 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: util.c
 *
 * Description: Utility functions for Rover server.
 *
 *
 * Global Functions: See HTML documentation.
 * Global Variables: See HTML documentation.
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include "rover.h"
#include "rovertcp.h"

struct timeval StartTv;
struct timezone Tz;

PUBLIC unsigned int ServerID;
PUBLIC Buffers *BufferHead = NULL;

PRIVATE char logbuf[1024];
PRIVATE Buffers *FindBuffer(int size, int allocate);

/*
 * Copy a string.
 */
PUBLIC char *copyString(char *string)
{
  char *rv = strdup(string);

  if (rv == NULL) fatalError("Server unable to allocate memory buffer");
  return(rv);
}

/*
 * Initialize an output buffer.
 */
PUBLIC void outstrInit(int size)
{
  state->LogSize = state->LogCount = state->EmittingLog = state->EmittingData =
    state->OutSize = state->CodeSize = state->DVSize = state->DataSize = 0;
  if (state->output && (state->output->size < size))
    outstrReleaseBuffer(state->output);
  state->output = FindBuffer(size, 1);
  state->output->buffer[0] = '\0';
}


/*
 * Find a matching buffer.
 */ 
PRIVATE Buffers *FindBuffer(int size, int allocate)
{
  Buffers *ptr = BufferHead, *prev = NULL;

  while (ptr) {
    if (ptr->size >= size) {
      state->output = ptr;
      if (prev) 
	prev->next = ptr->next;
      else
	BufferHead = ptr->next;
      if (RoverDebug)
	fprintf(stderr, "Reusing output buffer [FindBuffer:util.c]\n");
      return(ptr);
    }
    prev = ptr;
    ptr = ptr->next;
  }
 
  if (!allocate) return(NULL);

  if (RoverDebug) 
    fprintf(stderr, "Allocating new output buffer [FindBuffer:util.c]\n");
 
 /* Allocate a new buffer */
  if ((ptr = (Buffers *) malloc(sizeof(Buffers))) == NULL)
    fatalError("Unable to allocate output buffer unit");
  ptr->size = size;
  if ((ptr->buffer = malloc(size)) == NULL)
    fatalError("Unable to allocate output buffer [FindBuffer:util.c]");
  return(ptr);
}

/*
 * Release an output buffer.
 */ 
PUBLIC void outstrReleaseBuffer(Buffers *output)
{
  Buffers *ptr = BufferHead, *prev = NULL;
  int size = output->size;

  /* Insert the buffer sorted smallest to largest */
  while (ptr) {
    if (ptr->size > size) {
      output->next = ptr;
      if (prev) 
	prev->next = output;
      else
	BufferHead = output;
      return;
    }
    prev = ptr;
    ptr = ptr->next;
  }

  /* Add to the tail */
  output->next = NULL;
  if (prev) 
    prev->next = output;
  else
    BufferHead = output;
}

/*
 * Resize the output buffer.
 */
PUBLIC void outstrResize(int size)
{
  Buffers *ptr;
  int AllocSize = state->output->size;

  if (size < AllocSize) return;

  do {
    AllocSize <<= 1;
  } while (size >= AllocSize);

  /* First, try to find a matching buffer */
  ptr = FindBuffer(AllocSize, 0);
  if (ptr) {
    memcpy(ptr->buffer, state->output->buffer, state->OutSize);
    outstrReleaseBuffer(state->output);
    state->output = ptr;
  } else {
    state->output->buffer = realloc(state->output->buffer, AllocSize);
    state->output->size = AllocSize;
  }
}

/*
 * Add count bytes of string to the output buffer.
 */
PRIVATE inline void outstr(char *string, int count)
{
  int j = state->OutSize + count;
  if (j >= state->output->size) outstrResize(j);
  memcpy(state->output->buffer+state->OutSize, string, count);
  state->OutSize = j;
}

/*
 * Switch modes to emitting data.
 */
PUBLIC void emitData()
{
  if (state->EmittingLog) {
    state->status = 500;
    errorExit("Tried to emit data while emitting log");
  } else if (state->EmittingData) {
    state->status = 500;
    errorExit("Tried to emit data while already emitting data");
  }
  /* Allocate a placeholder for the code null */
  state->CodeSize++;
  outstr("X", 1);
  state->EmittingData = 1;
}

/*
 * Switch modes to emitting log records.
 */
PUBLIC void emitLog()
{
  if (state->EmittingLog) {
    state->status = 500;
    errorExit("Tried to emit log while already emitting log");
  } else if (!state->EmittingData) {
    /* Allocate a placeholder for the code null */
    state->CodeSize++;
    outstr("X", 1);
  }

  /* Allocate a placeholder for the data null */
  state->DataSize++;
  outstr("X", 1);
  state->EmittingLog = 1;
  state->LogStart=state->output->buffer+state->OutSize;/* Save ptr to log cnt*/
  sprintf(logbuf, "%7d\n", 0);
  outstr(logbuf, 8);
  state->LogSize = 8;
}

/*
 * Output a string in the data segment.
 */
PUBLIC void outstrD(char *string)
{
  int i = strlen(string);
  if (!state->EmittingData) {
    state->status = 500;
    errorExit("Tried to emit data while emitting code");
  } else if (state->EmittingLog) {
    state->status = 500;
    errorExit("Tried to emit data while emitting log");
  }
  state->DataSize += i;
  outstr(string, i);
}

/*
 * Output a string in the code segment.
 */
PUBLIC void outstrC(char *string)
{
  int i = strlen(string);
  if (state->EmittingData) {
    state->status = 500;
    errorExit("Tried to emit code while emitting data");
  } else if (state->EmittingLog) {
    state->status = 500;
    errorExit("Tried to emit code while emitting log");
  }
  state->CodeSize += i;
  outstr(string, i);
}

/*
 * Output a log record in the log segment.
 */
PUBLIC void outstrL(time_t timestamp, char *host, int logID, RoverOpStatus status,
		    char *operation)
{
  int len, oplen;

  if (state->EmittingData) {
    state->status = 500;
    errorExit("Tried to emit log while emitting data");
  } else if (!state->EmittingLog) {
    state->status = 500;
    errorExit("Tried to emit log while emitting code");
  }
  state->LogCount++;
  oplen = strlen(operation);
  sprintf(logbuf, "%d %s %d %d %d\n", (int) timestamp, host, logID, status,
	  oplen);
  len = strlen(logbuf);
  outstr(logbuf, len);
  outstr(operation, oplen);
  outstr("\n", 1);	/* Placeholder for null */
  state->LogSize += (len+oplen+1);
}

/*
 * Output count bytes of binary data.
 */
PUBLIC void outstrBin(void *bin, int count)
{
  if (state->EmittingData || state->CodeSize || state->EmittingLog) {
    state->status = 500;
    errorExit("Tried to emit binary data while emitting code, data, or log");
  }
  outstr((char *)bin, count);
}

/*
 * Output count bytes of string in the data segment.
 */
PUBLIC void outstrnD(char *string, int count)
{
  if (state->EmittingLog) {
    state->status = 500;
    errorExit("Tried to emit data while emitting log");
  } 
  state->DataSize += count;
  outstr(string, count);
}

/*
 * Output count bytes of string in the code segment.
 */
PUBLIC void outstrnC(char *string, int count)
{
  if (state->EmittingLog || state->EmittingData) {
    state->status = 500;
    errorExit("Tried to emit code while emitting data or log");
  } 
  state->CodeSize += count;
  outstr(string, count);
}

/*
 * Output text in the data segement, applying substitutions first.
 *   Count specifies the number of values (up to 9) provided for
 *   substitution in template. Each substition within template is
 *   specified as %n, where n refers to the nth provided value.
 */
PUBLIC void outstrVD(char *template, int count, ...)
{
  va_list ap;
  char *ptr, *string[10];
  int i, c;

  if ((count > 9) || (count < 1)) {
    fprintf(stderr, "outstrVD called with bad number of args %d", count);
    state->status = 500;
    errorExit("outstrVD called with bad number of args ");
  }

#ifdef ultrix
  va_start(ap);
#else
  va_start(ap, count);
#endif /* ultrix */
  for (i = 0; i < count; i++)
    string[i] = va_arg(ap, char *);

  while (*template) {
    ptr = index(template, '%');
    if (ptr == NULL) {				/* Any % chars left? */
      outstrD(template);
      return;
    } else {
      c = (int) ((*(ptr+1) - '0') & 0xFF);
      if ((c < 0) || (c >= count)) {		/* Is this just a %? */
	outstrnD(template,ptr-template+1); 	/* Copy the string */
	template = ptr+1;			/* Start all over again */
      } else {
	outstrnD(template, ptr-template);	/* Copy the string */
	outstrD(string[c]);			/* Send the name */
	template = ptr + 2;			/* Skip %# */
      }
    }
  }
  va_end(ap);
}

/*
 * Output text in the code segement, applying substitutions first.
 *   Count specifies the number of values (up to 9) provided for
 *   substitution in template. Each substition within template is
 *   specified as %n, where n refers to the nth provided value.
 */
PUBLIC void outstrVC(char *template, int count, ...)
{
  va_list ap;
  char *ptr, *string[10];
  int i, c;

  if ((count > 9) || (count < 1)) {
    fprintf(stderr, "outstrVC called with bad number of args %d", count);
    state->status = 500;
    errorExit("outstrVC called with bad number of args ");
  }

#ifdef ultrix
  va_start(ap);
#else
  va_start(ap, count);
#endif /* ultrix */
  for (i = 0; i < count; i++)
    string[i] = va_arg(ap, char *);

  while (*template) {
    ptr = index(template, '%');
    if (ptr == NULL) {				/* Any % chars left? */
      outstrC(template);
      return;
    } else {
      c = (int) ((*(ptr+1) - '0') & 0xFF);
      if ((c < 0) || (c >= count)) {		/* Is this just a %? */
	outstrnC(template,ptr-template+1); 	/* Copy the string */
	template = ptr+1;			/* Start all over again */
      } else {
	outstrnC(template, ptr-template);	/* Copy the string */
	outstrC(string[c]);			/* Send the name */
	template = ptr + 2;			/* Skip %# */
      }
    }
  }
  va_end(ap);
}

/*
 * Return a printable version of the time specified by stamp.
 */
PUBLIC char *rover_ctime(time_t *stamp)
{
  char *rv = ctime(stamp);
  rv[strlen(rv)-1] = '\0';	/* Trim off the \n */
  return(rv);
}

/*
 * Return the CMDtype enumeration value associated with the name.
 */
PUBLIC CMDtype Name2Cmd(char *name)
{
  if (!strcmp(name, ROVER_IMPORTBINARY_CMD)) return(IMPORTBINARY);
  if (!strcmp(name, ROVER_IMPORTDATA_CMD)) return(IMPORTDATA);
  if (!strcmp(name, ROVER_IMPORT_CMD)) return(IMPORTCMD);
  if (!strcmp(name, ROVER_EXPORT_CMD)) return(EXPORTCMD);
  if (!strcmp(name, ROVER_HTTP_CMD)) return(HTTP);
  if (!strcmp(name, ROVER_TEST_CMD)) return(TEST);
  if (!strcmp(name, ROVER_RESET_CMD)) return(RESET);
  if (!strcmp(name, ROVER_SMTP_CMD)) return(SMTP);
  return(NONE);
}

/*
 * Return the name associated with a CMDtype enumeration value.
 */
PUBLIC char *Cmd2Name(CMDtype cmd)
{
  switch(cmd) {
  case IMPORTBINARY:
    return(ROVER_IMPORTBINARY_CMD);
    break;
  case IMPORTDATA:
    return(ROVER_IMPORTDATA_CMD);
    break;
  case IMPORTCMD:
    return(ROVER_IMPORT_CMD);
    break;
  case EXPORTCMD:
    return(ROVER_EXPORT_CMD);
    break;
  case TEST:
    return(ROVER_TEST_CMD);
    break;
  case SMTP:
    return(ROVER_SMTP_CMD);
    break;
  default:
    return(NULL);
  }
}

/*
 * Set the server's unique server ID.
 */
PUBLIC void set_serverID()
{
#include "RoverServerID.h"
}

/*
 * Invoked when the serve encounters a recoverable error.
 *   Returns an error object to the user.
 */
PUBLIC void errorExit(char *msg)
{
  int rc, extra = 0;

  fprintf(stderr, "Rover server called ErrorExit with %d (%s)\n",
	  state->status, msg);
  if (!state->CommandDone) {
    state->status = 200;/* Return encoded error object with a good HTTP code */
    outstrInit(INITSIZE);
    emitData();
    outstrD(msg);
    /* Allocate a placeholder for the data null */
    outstrD("X");
    state->ObjType = ROVER_ERROR_OBJ;
    /*
     * Mark the object as best effort so that it will be cached (in case the 
     * client is disconnected and retries the action), but the client will
     * resend the operation if it is connected).
     */
    state->Cacheable = ROVER_BEST_EFFORT;
    state->DVSize = 0;
    state->DVbuf[0] = '\0';
    rc = sendServerResponse(&extra, 0);
    fprintf(stderr, "errorExit: Actually wrote %d+0%d bytes\n", rc, extra);
  }
  globalTclInterp = NULL;
  state->auth.Nobody = 0;

  /* Need to deal with what happens when this isn't set */
  longjmp(state->RoverEnv, ROVER_ERROR);
}

/* 
 * Fatal error exit.
 *   Invoked when the server has encountered an unrecoverable error.
 */
PUBLIC void fatalError(char *msg)
{
  fprintf(stderr, "Rover server - Unrecoverable fatal error: %s\n", msg);

  /* Need to deal with what happens when this isn't set */
  longjmp(state->RoverEnv, ROVER_FATAL);
}

PUBLIC void testExit(int size)
{
  long diff;
  struct timeval EndTv;
  struct timezone EndTz;
  int s = size;

  fputc('\n', state->Stdout);
  while (size--)
    fputc('a', state->Stdout);
  fputc('\n', state->Stdout);
  fflush(state->Stdout);
  gettimeofday(&EndTv, &EndTz);
  diff = (EndTv.tv_sec - StartTv.tv_sec) * 1000000;	/* Sec -> us */
  diff += (EndTv.tv_usec - StartTv.tv_usec);
  fprintf(stderr, "TEST: Size %d Time: %.3f\n", s, ((double)diff)/1e6);
  state->CommandDone = 1;
}


PRIVATE char Name[300];

PUBLIC int stableHashInit(char *tableName)
{
  Tcl_HashTable *table = NULL;
  int new, create = 0;

  sprintf(Name, "Hash_%s", tableName);
  if ((hentry = Tcl_FindHashEntry(&Info, Name)) != NULL) {
    table = (Tcl_HashTable *)Tcl_GetHashValue(hentry);

#if 0
    if (!state->msgID) {
      /* Message series reset, delete the hash table */
      fprintf(stderr, "Message series reset, deleting hash table: %s\n",
	      tableName);
      Tcl_DeleteHashTable(table);
      table = NULL;
    }
#endif

  }
  if (!table) {
    create = 1;
    if ((table = (Tcl_HashTable *)malloc(sizeof(Tcl_HashTable))) == NULL) {
      fatalError("Server unable to allocate hash table");
    }
    Tcl_InitHashTable(table, TCL_STRING_KEYS);
    hentry = Tcl_CreateHashEntry(&Info, Name, &new);
    Tcl_SetHashValue(hentry, (char *)table);	/* Save it */
    if (standalone && (ServerRecovering != READLOG) && !ServerNoLog) 
      fprintf(ServerLogFile, "ROVER_LOG_STABLE_HTDCL %s\n", tableName);
  }
  return(create);
}

PUBLIC char *stableHashFetch(char *tableName, char *entryName)
{
  Tcl_HashTable *table;

  sprintf(Name, "Hash_%s", tableName);
  if ((hentry = Tcl_FindHashEntry(&Info, Name)) != NULL) {
    table = (Tcl_HashTable *)Tcl_GetHashValue(hentry);
    if ((hentry = Tcl_FindHashEntry(table, entryName)) != NULL) 
      return(((HashEntry *)Tcl_GetHashValue(hentry))->data);
  }
  return(NULL);
}

PUBLIC int stableHashSetEntry(char *tableName, char *entryName, char *data,
			      int size, char **old)
{
  Tcl_HashTable *table;
  int new;
  HashEntry *entry;

  *old = NULL;
  if (index(entryName, ' ')) {
    fprintf(stderr, "Error! Hash table entry name `%s' contains space(s)\n",
	    entryName);
    return(-1);
  }
  sprintf(Name, "Hash_%s", tableName);
  if ((hentry = Tcl_FindHashEntry(&Info, Name)) != NULL) 
    table = (Tcl_HashTable *)Tcl_GetHashValue(hentry);
  else  {
    fprintf(stderr, "Hash table `%s' not found\n", tableName);
    return(-1);
  }

  hentry = Tcl_CreateHashEntry(table, entryName, &new);
  if (!new) {				/* Check for duplicates */
    fprintf(stderr,"Updating existing entry in hash table %s: %s\n", tableName,
	    entryName);
    entry = (HashEntry *)Tcl_GetHashValue(hentry);
    *old = entry->data;
  } else {
    if ((entry = (HashEntry *)malloc(sizeof(HashEntry))) == NULL) {
      fatalError("Server unable to allocate hash table entry");
    }
    Tcl_SetHashValue(hentry, (char *) entry);
  }
  entry->data = data;
  entry->size = size;

  fprintf(stderr, "Rover stable hash table %s entry %s changed: %s\n",
	  tableName, entryName, data);

  if (!standalone || (ServerRecovering == READLOG) || ServerNoLog) return(0);
  fprintf(ServerLogFile, "ROVER_LOG_STABLE_HTVAR %s %s %d\n", tableName,
	  entryName, size);
  if (size) {
    if ((new = fwrite(data, sizeof(char), size, ServerLogFile)) != size) {
      state->status = 500;
      sprintf(logbuf,"Bad write size %d [stableHashSetEntry:util.c]\n",new );
      errorExit(logbuf);
    }
  }
  fputc('\n', ServerLogFile);
  fflush(ServerLogFile);
  if (ServerHardFlush) fsync(ServerLogFD);
  return(0);
}

PUBLIC void stableLogOutput()
{
  int rc, total = (state->OutSize + state->DVSize);

  if (!standalone || (ServerRecovering == READLOG) || ServerNoLog) return;
  state->loggedOutput = 2;
  fprintf(ServerLogFile2, "ROVER_LOG_OUTPUT %d %d %d %d %d %d %d %d\n",
	  state->ServerSeq, state->ObjType, state->CodeSize, state->DataSize,
	  state->Cacheable, state->LogSize, state->DVSize, total);
  if (total) {
    if ((rc = fwrite(state->output->buffer, sizeof(char), state->OutSize,
		     ServerLogFile2))!= rc) {
      state->status = 500;
      sprintf(logbuf,"Bad buf write size %d [stableLogOutput:util.c]\n",rc );
      errorExit(logbuf);
    }
    if ((rc = fwrite(state->DVbuf, sizeof(char),state->DVSize,ServerLogFile2))
	!= rc) {
      state->status = 500;
      sprintf(logbuf,"Bad dv write size %d [stableLogOutput:util.c]\n",rc );
      errorExit(logbuf);
    }
  }
  fputc('\n', ServerLogFile2);
  fflush(ServerLogFile2);
  if (ServerHardFlush) fsync(ServerLogFD2);
}


PUBLIC void roverResetAuth(ServerState *s)
{
  s->auth.Nobody = s->auth.valid = 0;
  s->auth.gid = -1;
  s->auth.uid = -1;
  if (s->auth.HomeEnv) {
    free(s->auth.HomeEnv);
    s->auth.HomeEnv = NULL;
  }
  if (s->auth.UserEnv) {
    free(s->auth.UserEnv);
    s->auth.UserEnv = NULL;
  }
}

PUBLIC int roverInitLogFile(char *name, FILE **filep, int *fdp)
{
  fprintf(stderr, "Initializing log file: %s\n", name);
  *filep = fopen(name, "a");
  if (*filep == NULL) {
    fprintf(stderr, "Open of file %s failed [roverInitLogFile:util.c]\n",name);
    return(-1);
  }
  *fdp = fileno(*filep);
  fprintf(*filep, "ROVER_LOG_VERSION %lu\n",
	  (unsigned long) ROVER_SERVER_LOG_VERSION);
  fflush(*filep);
  return(0);
}

PUBLIC void roverCloseLogs()
{
  fflush(ServerLogFile);
  fclose(ServerLogFile);

  fflush(ServerLogFile2);
  fclose(ServerLogFile2);
}

PUBLIC int checkVerify(char *object, ServerState *s, RoverObjType utype)
{
  if (s->dvtime <= s->UserTime) {
    /* We've verified that the object hasn't changed at the server */
    if (RoverDebug)
      fprintf(stderr, "Verified object `%s', returning header\n", object);
    s->Cacheable = ROVER_VERIFIED;
    s->compress = 0;
    return(1);
  } else if (s->UserTime) {
    if (RoverDebug)
      fprintf(stderr, "Updating client's copy of object `%s'\n", object);
    s->ObjType = utype;
  }
  if (RoverDebug)
    fprintf(stderr, "Object `%s' has time %s (%lu) vs. user time %s (%lu)\n",
	    object, rover_ctime(&(s->dvtime)), s->dvtime,
	    rover_ctime(&(s->UserTime)), s->UserTime);
  return(0);
}

